package org.artifactory.event.priority.service;

import org.artifactory.event.priority.service.model.PrioritizedNodeEvent;
import org.artifactory.event.priority.service.model.PrioritizedNodeEventMapper;
import org.artifactory.event.priority.service.strategy.BasePriorityChartStrategy;
import org.artifactory.event.priority.service.strategy.ExecutionPredicatePriorityChartStrategy;
import org.artifactory.event.priority.service.strategy.IdenticalPathChartStrategy;
import org.artifactory.event.priority.service.strategy.PriorityChartStrategy;
import org.artifactory.event.provider.EventProvider;
import org.artifactory.event.work.NodeEventOperator;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.db.event.dao.NodeEventPrioritiesDao;
import org.artifactory.storage.db.event.model.NodeEventCursor;
import org.artifactory.storage.db.event.service.NodeEventCursorService;
import org.artifactory.storage.event.NodeEvent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * @author Uriah Levy
 */
@Service
public class NodeEventPriorityServiceImpl implements NodeEventPriorityService {
    private static final Logger log = LoggerFactory.getLogger(NodeEventPriorityServiceImpl.class);
    private NodeEventCursorService eventCursorService;
    private NodeEventPriorityStorageService eventPriorityStorageService;
    private PrioritizedNodeEventMapper prioritizedEventMapper;
    private NodeEventPrioritiesDao eventPrioritiesDao;

    private List<PriorityChartStrategy> priorityChartStrategyList = new ArrayList<>();
    private static final int MAX_RETRIES = 5;

    @Autowired
    public NodeEventPriorityServiceImpl(NodeEventCursorService eventCursorService,
            NodeEventPriorityStorageService eventPriorityStorageService, NodeEventPrioritiesDao prioritiesDao,
            PrioritizedNodeEventMapper mapper) {
        this.eventCursorService = eventCursorService;
        this.eventPriorityStorageService = eventPriorityStorageService;
        this.eventPrioritiesDao = prioritiesDao;
        this.prioritizedEventMapper = mapper;
        priorityChartStrategyList.add(new BasePriorityChartStrategy());
        priorityChartStrategyList.add(new IdenticalPathChartStrategy());
    }

    @Override
    public void createAndSaveEventPrioritiesIfNeeded(NodeEventOperator operator, EventProvider<? extends NodeEvent> eventProvider) {
        NodeEventPriorityChart priorityChart = NodeEventPriorityChartHelper
                .chartFromEvents(getNextBatch(operator, eventProvider), operator.getOperatorId());
        if (!priorityChart.isEmpty()) {
            final long finalTimestamp = priorityChart.getFinalTimestamp();
            applyPriorityStrategies(priorityChart, operator);
            saveChartAndUpdateCursor(operator, priorityChart, finalTimestamp);
        }
    }

    @Override
    public List<PrioritizedNodeEvent> getCurrentPriorityBatch(String operatorID) {
        try {
            Optional<Integer> minimalPriorityByOperatorId = eventPrioritiesDao
                    .getMinimalPriorityByOperatorId(operatorID);
            if (minimalPriorityByOperatorId.isPresent() && minimalPriorityByOperatorId.get() > 0) {
                log.debug("Current minimal priority is {}", minimalPriorityByOperatorId.get());
                return eventPrioritiesDao
                        .getEventsByPriorityAndOperatorId(minimalPriorityByOperatorId.get(), operatorID)
                        .stream()
                        .filter(prioritizedEvent -> prioritizedEvent.getRetryCount() < MAX_RETRIES)
                        .map(prioritizedEventMapper::toEvent)
                        .collect(Collectors.toList());
            } else {
                log.debug("No prioritized node events were found");
            }
        } catch (SQLException e) {
            throw new StorageException("Unable to save new event priority", e);
        }
        return Collections.emptyList();
    }

    @Override
    public void deleteEventPriority(PrioritizedNodeEvent prioritizedNodeEvent) {
        try {
            eventPrioritiesDao.deleteEventPriorityById(prioritizedNodeEvent.getPriorityId());
        } catch (SQLException e) {
            throw new StorageException(
                    "Unable to delete event priority with ID " + prioritizedNodeEvent.getPriorityId(), e);
        }
    }

    @Override
    public boolean priorityListIsEmpty(String operatorId) {
        try {
            Optional<Integer> minimalPriorityByOperatorId = eventPrioritiesDao
                    .getMinimalPriorityByOperatorId(operatorId);
            if (minimalPriorityByOperatorId.isPresent()) {
                // The minimal ordinal priority is 1. Zero means the DB has no priorities for this operator
                return minimalPriorityByOperatorId.get() == 0;
            }
        } catch (SQLException e) {
            log.error("Unable to determine whether priority list is empty for operator with ID", e);
        }
        return false;
    }

    @Override
    public int getNumberOfPrioritizedEvents(String operatorId) {
        try {
            return eventPrioritiesDao.countRetryableEventsByOperatorId(operatorId).orElse(0);
        } catch (SQLException e) {
            throw new StorageException("Unable to count events for operator with ID " + operatorId);
        }
    }

    @Override
    public void saveRetryCount(PrioritizedNodeEvent prioritizedNodeEvent) {
        try {
            int retryCount = prioritizedNodeEvent.getRetryCount();
            log.debug("Incrementing retry count on event with path '{}' to {}", prioritizedNodeEvent.getPath(),
                    retryCount);
            eventPrioritiesDao.updateRetryCountById(prioritizedNodeEvent.getPriorityId(), retryCount);
            if (retryCount == MAX_RETRIES) {
                log.warn("Retry attempts exhausted at '{}' retries for '{}'", retryCount,
                        prioritizedNodeEvent.getPath());
                deleteEventPriority(prioritizedNodeEvent);
            }
        } catch (SQLException e) {
            throw new StorageException(
                    "Unable to bump retry count on event with path " + prioritizedNodeEvent.getPath());
        }
    }

    private <T extends NodeEvent> List<T> getNextBatch(NodeEventOperator operator, EventProvider<T> eventProvider) {
        NodeEventCursor lastOperatorCursor = eventCursorService.cursorForOperator(operator.getOperatorId());
        if (lastOperatorCursor != null) {
            return eventProvider.provideNextBatch(lastOperatorCursor);
        }
        return Collections.emptyList();
    }

    private void applyPriorityStrategies(NodeEventPriorityChart nodeEventPriorityChart, NodeEventOperator operator) {
        ExecutionPredicatePriorityChartStrategy operatorExecutionStrategy = new ExecutionPredicatePriorityChartStrategy(
                operator.getExecutionPredicate());
        operatorExecutionStrategy.apply(nodeEventPriorityChart);
        priorityChartStrategyList.forEach(strategy -> strategy.apply(nodeEventPriorityChart));
    }

    private void saveChartAndUpdateCursor(NodeEventOperator operator,
            NodeEventPriorityChart nodeEventPriorityChart, long finalTimestamp) {
        if (!nodeEventPriorityChart.isEmpty()) {
            eventPriorityStorageService.savePriorityChart(nodeEventPriorityChart);
        }
        eventCursorService.updateCursor(new NodeEventCursor(operator.getOperatorId(), finalTimestamp));
    }
}
