package org.artifactory.event.work;

import com.google.common.util.concurrent.ThreadFactoryBuilder;
import org.artifactory.event.priority.service.NodeEventPriorityService;
import org.artifactory.event.provider.EventProvider;
import org.artifactory.storage.event.NodeEvent;
import org.jfrog.storage.common.ConflictGuard;
import org.jfrog.storage.common.ConflictsGuard;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;

class EventOperatorDispatcher extends Thread {
    private static final Logger log = LoggerFactory.getLogger(EventOperatorDispatcher.class);
    private static final int MAX_BATCH_ERRORS = 50;
    private NodeEventPriorityService nodeEventPriorityService;
    private EventProvider<? extends NodeEvent> eventProvider;
    private NodeEventOperator nodeEventOperator;
    private boolean active = true;
    private ConflictsGuard<String> conflictsGuard;
    private int maxPermits;
    private final Semaphore semaphore;
    private final Object workMonitor = new Object();
    private AtomicInteger batchErrors = new AtomicInteger();
    private ExecutorService executor;

    EventOperatorDispatcher(NodeEventPriorityService nodeEventPriorityService, EventProvider<? extends NodeEvent> eventProvider,
            NodeEventOperator operator,
            ConflictsGuard<String> conflictsGuard, int numOfThreads) {
        this.nodeEventPriorityService = nodeEventPriorityService;
        this.eventProvider = eventProvider;
        this.nodeEventOperator = operator;
        this.conflictsGuard = conflictsGuard;
        this.maxPermits = numOfThreads;
        this.semaphore = new Semaphore(numOfThreads);
        ThreadFactory threadFactory = new ThreadFactoryBuilder()
                .setNameFormat(operator.getOperatorId() + "-sweeper-%s").setDaemon(true).build();
        executor = new ThreadPoolExecutor(numOfThreads, numOfThreads, 5L, TimeUnit.MINUTES, new LinkedBlockingQueue<>(),
                threadFactory);
    }

    @Override
    public void run() {
        try {
            dispatchLoop();
        } finally {
            stopSweepers();
        }
    }

    void batchError() {
        if (batchErrors.incrementAndGet() >= MAX_BATCH_ERRORS) {
            log.warn("Event error threshold reached for operator '{}'", nodeEventOperator.getOperatorId());
            nodeEventOperator.onErrorThresholdReached();
        }
    }

    /**
     * Creates the prioritized list of events, and dispatches work sweepers in a semi back-pressured sort of way.
     */
    private void dispatchLoop() {
        synchronized (workMonitor) {
            while (active) {
                try {
                    if (nodeEventOperator.isEnabled()) {
                        createPrioritiesIfNeeded();
                        submitSweepers();
                    }
                    // Wait for workers to signal work done
                    workMonitor.wait(TimeUnit.SECONDS.toMillis(10));
                } catch (InterruptedException ie) {
                    Thread.currentThread().interrupt();
                } catch (Exception e) {
                    log.warn("Caught exception during the '{}' dispatch loop. The dispatcher will continue execution",
                            nodeEventOperator.getOperatorId());
                    try {
                        workMonitor.wait(TimeUnit.SECONDS.toMillis(10));
                    } catch (InterruptedException e1) {
                        Thread.currentThread().interrupt();
                    }
                }
            }
        }
    }

    private void submitSweepers() {
        int activeWorkers = maxPermits - semaphore.availablePermits();
        while (nodeEventPriorityService.getNumberOfPrioritizedEvents(nodeEventOperator.getOperatorId()) >
                activeWorkers &&
                semaphore.tryAcquire()) {
            submitSweeper();
        }
    }

    void submitSweeper() {
        executor.submit(
                new NodeEventSweeper(nodeEventPriorityService, nodeEventOperator, this));
    }

    private void createPrioritiesIfNeeded() {
        if (semaphore.availablePermits() == maxPermits) {
            try {
                if (nodeEventPriorityService.priorityListIsEmpty(nodeEventOperator.getOperatorId())) {
                    runExclusive(getLockKeyForOperator(), () -> doCreatePriorities(nodeEventOperator.getOperatorId()));
                }
            } catch (Exception e) {
                log.error("Caught exception while creating the prioritized event list.", e);
            }
        }
    }

    private void doCreatePriorities(String operatorId) {
        if (nodeEventPriorityService.priorityListIsEmpty(operatorId)) {
            log.debug("Creating new priority list for operator '{}'", operatorId);
            nodeEventPriorityService.createAndSaveEventPrioritiesIfNeeded(nodeEventOperator, eventProvider);
            batchErrors = new AtomicInteger();
        }
    }


    void runExclusive(String lockKey, Runnable task) {
        ConflictGuard lock = conflictsGuard.getLock(lockKey);
        boolean lockAcquired = false;
        try {
            lockAcquired = lock.tryToLock(0, TimeUnit.SECONDS);
            if (lockAcquired) {
                log.debug("Lock acquired on key {}", lockKey);
                task.run();
                return;
            }
            log.debug("key '{}' is already locked.", lockKey);
        } catch (InterruptedException e) {
            log.error("Node Event Manager dispatcher or worker thread interrupted.", e);
            Thread.currentThread().interrupt();
        } finally {
            releaseLock(lockKey, lock, lockAcquired);
        }
    }

    private void releaseLock(String lockKey, ConflictGuard lock, boolean lockAcquired) {
        try {
            if (lockAcquired) {
                lock.unlock();
                log.debug("Lock on key '{}' released in finally block", lockKey);
            }
        } catch (Exception e) {
            log.error("Unable to release lock with key {}", lockKey, e);
        }
    }

    private String getLockKeyForOperator() {
        return nodeEventOperator.getOperatorId() + "-priorities";
    }

    void stopDispatcher() {
        active = false;
    }

    private void stopSweepers() {
        try {
            executor.awaitTermination(5, TimeUnit.SECONDS);
            executor.shutdownNow();
        } catch (InterruptedException e) {
            log.error("Unable to stop dispatched sweepers", e);
            Thread.currentThread().interrupt();
        }
    }

    /**
     * Called by the sweepers once (and only once), when a sweeper finishes working.
     */
    void sweeperDone() {
        synchronized (workMonitor) {
            workMonitor.notifyAll();
        }
        semaphore.release();
    }
}