package org.artifactory.event.work;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Supplier;

import static java.lang.Thread.currentThread;
import static java.lang.Thread.sleep;

/**
 * @author Uriah Levy
 */
@Component
public class StatusProbeServiceImpl implements StatusProbeService {
    private static final Logger log = LoggerFactory.getLogger(StatusProbeServiceImpl.class);

    private long intervalTimeMillis = TimeUnit.SECONDS.toMillis(45);
    private int maxAttempts = 2000;
    private ReentrantLock probingLock = new ReentrantLock();

    @Override
    public void startProbing(ExecutorService executor, Supplier<Boolean> probeFunction, Runnable onSuccess) {
        executor.submit(() -> locallyExclusive(() -> {
            boolean probeSuccess = false;
            int currentAttempt = 0;
            while (!probeSuccess && currentAttempt < maxAttempts) { // 45 * 2000 = 25 hours of retries
                try {
                    probeSuccess = probeFunction.get();
                    currentAttempt++;
                    if (!probeSuccess) {
                        pause();
                    }
                } catch (Exception e) {
                    currentAttempt++;
                    pause();
                }
            }
            // Success
            onSuccess.run();
        }));
    }

    @Override
    public boolean isProbing() {
        return probingLock.isLocked();
    }

    private void pause() {
        try {
            log.debug("Probing function indicated failure. Waiting for {} seconds before retrying...",
                    intervalTimeMillis);
            sleep(intervalTimeMillis);
        } catch (InterruptedException e) {
            log.warn("Probing function interrupted", e);
            currentThread().interrupt();
            throw new IllegalStateException("Probing function interrupted.", e);
        }
    }

    private void locallyExclusive(Runnable function) {
        try {
            boolean lockAcquired = probingLock.tryLock();
            if (lockAcquired) {
                log.debug("Local probing exclusive lock acquired");
                function.run();
            }
        } finally {
            if (probingLock.isHeldByCurrentThread()) {
                probingLock.unlock();
                log.debug("Local probing exclusive lock released");
            }
        }
    }
}
