package org.artifactory.metadata.service;

import com.google.common.util.concurrent.MoreExecutors;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang3.mutable.MutableBoolean;
import org.apache.commons.lang3.tuple.Pair;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.aql.AqlService;
import org.artifactory.common.ConstantValues;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.event.EventOperatorId;
import org.artifactory.event.priority.service.model.PrioritizedNodeEvent;
import org.artifactory.event.work.NodeEventOperator;
import org.artifactory.event.work.NodeEventTaskManager;
import org.artifactory.event.work.StatusProbeService;
import org.artifactory.fs.FileInfo;
import org.artifactory.fs.ItemInfo;
import org.artifactory.fs.StatsInfo;
import org.artifactory.md.Properties;
import org.artifactory.metadata.service.provider.AbstractMetadataProvider;
import org.artifactory.metadata.service.store.ArtifactoryMetadataClientConfigStore;
import org.artifactory.repo.RepoPath;
import org.artifactory.security.access.AccessService;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.spring.Reloadable;
import org.artifactory.storage.db.DbService;
import org.artifactory.storage.db.event.model.NodeEventCursor;
import org.artifactory.storage.db.event.service.NodeEventCursorService;
import org.artifactory.storage.db.event.service.metadata.mapper.MetadataEntityMapper;
import org.artifactory.storage.db.event.service.metadata.model.MetadataEvent;
import org.artifactory.storage.db.event.service.metadata.model.MutableMetadataEntityBOM;
import org.artifactory.storage.db.fs.dao.NodesDao;
import org.artifactory.storage.db.migration.service.MigrationStatusStorageService;
import org.artifactory.storage.event.EventType;
import org.artifactory.storage.event.EventsService;
import org.artifactory.storage.fs.service.FileService;
import org.artifactory.storage.fs.service.PropertiesService;
import org.artifactory.storage.fs.service.StatsService;
import org.jfrog.metadata.client.MetadataClient;
import org.jfrog.metadata.client.confstore.MetadataClientConfigStore;
import org.jfrog.metadata.client.model.*;
import org.jfrog.metadata.client.model.event.MetadataEventEntity;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;

import javax.annotation.Nonnull;
import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Collectors;

/**
 * A business service that handles Metadata Events. Responsible for starting the event pipeline and initializing
 * the {@link MetadataClient}. Additionally, this service implements the {@link NodeEventOperator} which provides
 * the logic for handling Metadata events.
 *
 * @author Uriah Levy
 */
@Reloadable(beanClass = MetadataEventService.class, initAfter = {DbService.class, AccessService.class})
public class MetadataEventServiceImpl implements MetadataEventService, ContextCreationListener {

    private static final Logger log = LoggerFactory.getLogger(MetadataEventServiceImpl.class);

    private Map<RepoType, AbstractMetadataProvider> metadataProvidersMap;
    private RepositoryService repositoryService;
    private FileService fileService;
    private PropertiesService propertiesService;
    private NodeEventCursorService nodeEventCursorService;
    private AccessService accessService;
    private NodeEventTaskManager nodeEventTaskManager;
    private StatusProbeService statusProbeService;
    private MetadataEntityMapper metadataEntityMapper;
    private MetadataClient metadataClient;
    private MetadataClientConfigStore configStore;
    private boolean operatorEnabled = false;
    private ExecutorService asyncProbeExecutor;
    private ReentrantLock serviceStateLock = new ReentrantLock();
    private MigrationStatusStorageService migrationStatusStorageService;
    private NodesDao nodesDao;
    private AqlService aqlService;
    private EventsService eventsService;


    @Autowired
    public MetadataEventServiceImpl(List<AbstractMetadataProvider> metadataProviders,
            MetadataEntityMapper metadataEntityMapper) {
        this.metadataProvidersMap = metadataProviders.stream()
                .collect(Collectors.toMap(AbstractMetadataProvider::getRepoType,
                        Function.identity()));
        this.metadataEntityMapper = metadataEntityMapper;
        asyncProbeExecutor = new ThreadPoolExecutor(1, 1, 0, TimeUnit.SECONDS, new LinkedBlockingDeque<>(1),
                new ThreadFactoryBuilder().setDaemon(true)
                        .setNameFormat(getOperatorId() + "-status-probe").build());
    }

    @Override
    public void init() {
        initOperatorCursor();
        configStore = new ArtifactoryMetadataClientConfigStore(accessService);
        initMetadataClient();
    }

    @Override
    public void onContextCreated() {
        MetadataMigrationHelper metadataMigrationHelper = new MetadataMigrationHelper(migrationStatusStorageService,
                nodeEventCursorService, nodesDao, aqlService,
                eventsService);
        metadataMigrationHelper.migrateOrStartEventPipe(this, nodeEventTaskManager);
    }

    @Override
    public boolean isEnabled() {
        return this.operatorEnabled;
    }

    @Override
    public MetadataClient getMetadataClient() {
        return metadataClient;
    }

    @Override
    public void initOperatorCursor() {
        initEventCursorIfNeeded();
    }

    @Override
    public void operateOnEvent(PrioritizedNodeEvent nodeEvent) {
        MetadataEvent metadataEvent = prioritizedEventToMetadataEvent(nodeEvent);
        if (metadataEvent != null) {
            handleMetadataEvent(metadataEvent);
        }
    }

    public boolean meetsExecutionCriteria(PrioritizedNodeEvent nodeEvent) {
        Optional<AbstractMetadataProvider> metadataProvider = metadataProviderForItem(
                nodeEvent.getRepoPath().getRepoKey());
        if (!nodeEvent.getRepoPath().isFile() || nodeEvent.getRepoPath().isRoot()) { // <- root is a file :(
            return false;
        }
        MutableBoolean isLead = new MutableBoolean(false);
        metadataProvider.ifPresent(provider -> isLead.setValue(provider.isLeadArtifact(nodeEvent.getRepoPath())));
        return isLead.booleanValue();
    }

    @Override
    public String getOperatorId() {
        return EventOperatorId.METADATA_OPERATOR_ID.getId();
    }

    @Override
    public int getNumberOfEventThreads() {
        return ConstantValues.metadataEventOperatorThreads.getInt();
    }

    @Override
    public void onErrorThresholdReached() {
        disableServiceAndProbeSync();
    }

    @Override
    public Predicate<PrioritizedNodeEvent> getExecutionPredicate() {
        return this::meetsExecutionCriteria;
    }

    void setConfigStore(MetadataClientConfigStore configStore) {
        this.configStore = configStore;
    }

    void initMetadataClient() {
        closeCurrentClient();
        if (ConstantValues.metadataServerEventsEnabled.getBoolean()) { // silent feature flag
            try {
                metadataClient = configStore.authenticatedClientBuilder().create();
                operatorEnabled = metadataClient.system().ping();
            } catch (Exception e) {
                log.error("Unable to init the Metadata client. The Metadata Event pipeline will be disabled: {}",
                        e.getMessage());
                log.debug("", e);
                if (!statusProbeService.isProbing()) { // If we're already probing, avoid cyclic re-init
                    disableServiceAndProbeAsync();
                }
            }
        }
    }

    private void initEventCursorIfNeeded() {
        NodeEventCursor eventCursor = nodeEventCursorService.cursorForOperator(getOperatorId());
        if (eventCursor == null) {
            nodeEventCursorService.insertCursor(getOperatorId());
        }
    }

    private void handleMetadataEvent(MetadataEvent metadataEvent) {
        if (isPackageCreatedEvent(metadataEvent)) {
            createPackage(metadataEvent);
        } else if (isPathDeletedEvent(metadataEvent)) {
            deletePath(metadataEvent);
        }
    }

    private void createPackage(MetadataEvent metadataEvent) {
        MdsPackage mdsPackage = (MdsPackage) metadataEvent.getMetadataEntity();
        if (mdsPackage != null) {
            log.debug("Creating/updating Package Metadata for '{}' ({})", mdsPackage.getPkgid(),
                    metadataEvent.getPath());
            metadataClient.packages().create(mdsPackage);
        }
    }

    private void deletePath(MetadataEvent metadataEvent) {
        log.debug("Deleting Metadata Entity with path '{}'", metadataEvent.getPath());
        metadataClient.paths().deletePath((MdsVersionRepo) metadataEvent.getMetadataEntity());
    }

    private boolean isPathDeletedEvent(MetadataEvent metadataEvent) {
        return metadataEvent.getMetadataEventEntity().equals(MetadataEventEntity.PathDeleted);
    }

    private boolean isPackageCreatedEvent(MetadataEvent metadataEvent) {
        return metadataEvent.getMetadataEventEntity().equals(MetadataEventEntity.PackageCreated);
    }

    private void closeCurrentClient() {
        if (metadataClient != null) {
            IOUtils.closeQuietly(metadataClient);
        }
    }

    private void disableServiceAndProbeSync() {
        if (operatorEnabled) {
            this.operatorEnabled = false;
            statusProbeService
                    .startProbing(MoreExecutors.newDirectExecutorService(), this::probeMetadataServer,
                            this::onSuccessfulStatusProbe);
        }
    }

    private void disableServiceAndProbeAsync() {
        this.operatorEnabled = false;
        statusProbeService
                .startProbing(asyncProbeExecutor, this::probeMetadataServer, this::onSuccessfulStatusProbe);
    }

    private void onSuccessfulStatusProbe() {
        initMetadataClient();
        if (serviceStateLock.isHeldByCurrentThread()) {
            serviceStateLock.unlock();
        }
        log.info("Metadata Service Probing succeeded. The event pipeline was re-enabled.");
    }

    private boolean probeMetadataServer() {
        try (MetadataClient noAuthMetadataClient = configStore.noAuthClientBuilder().create()) {
            return noAuthMetadataClient.system().ping();
        } catch (IOException e) {
            log.error("Unable to create the status probe HTTP client");
            log.debug("Unable to create the status probe HTTP client", e);
        }
        return false;
    }

    ExecutorService getAsyncProbeExecutor() {
        return asyncProbeExecutor;
    }

    MetadataEvent prioritizedEventToMetadataEvent(PrioritizedNodeEvent prioritizedNodeEvent) {
        return createMetadataEntity(prioritizedNodeEvent)
                .map(entityToEventEntity -> createMetadataEvent(prioritizedNodeEvent, entityToEventEntity.getLeft(),
                        entityToEventEntity.getRight()))
                .orElse(null);
    }

    private Optional<Pair<MetadataEntity, MetadataEventEntity>> createMetadataEntity(PrioritizedNodeEvent event) {
        if (isDeleteEvent(event.getType())) {
            return Optional
                    .of(Pair.of(new MdsVersionRepo(event.getRepoPath().getRepoKey(), event.getRepoPath().getPath()),
                            MetadataEventEntity.PathDeleted));
        } else {
            Optional<ItemInfo> itemInfo = loadItem(event.getRepoPath());
            return itemInfo.map(item -> Pair.of(createMdsPackage(item), MetadataEventEntity.PackageCreated));
        }
    }

    private Optional<ItemInfo> loadItem(RepoPath repoPath) {
        try {
            return Optional.of(fileService.loadItem(repoPath));
        } catch (RuntimeException e) {
            log.debug("Unable to find item with path '{}'. Metadata event handling will be skipped.",
                    repoPath.toPath());
            return Optional.empty();
        }
    }

    private boolean isDeleteEvent(EventType type) {
        return type.equals(EventType.delete);
    }

    private MdsPackage createMdsPackage(ItemInfo item) {
        Properties properties = propertiesService.getProperties(item);
        StatsInfo stats = statsService().getStats((FileInfo) item);
        Optional<MutableMetadataEntityBOM> mutableMetadataBOM = createMutableMetadataEntityBOM(item, properties, stats);
        return mutableMetadataBOM
                .map(mutableMetadataEntityBOM -> metadataBomToMdsPackage(mutableMetadataEntityBOM, item.getRepoPath()))
                .orElse(null);
    }

    private MdsPackage metadataBomToMdsPackage(MutableMetadataEntityBOM mutableMetadataBOM,
            RepoPath repoPath) {
        MdsPackage mdsPackage = metadataEntityMapper.metadataBomToMdsPackage(mutableMetadataBOM);
        if (StringUtils.isBlank(mdsPackage.getPkgid())) {
            if (log.isDebugEnabled()) {
                log.debug("Ignoring metadata event with empty pkgid with path '{}'", repoPath.toPath());
            }
            return null;
        }
        addVersionFromBom(mutableMetadataBOM, mdsPackage);
        addFilesFromBom(mutableMetadataBOM, mdsPackage);
        return mdsPackage;
    }

    private void addFilesFromBom(MutableMetadataEntityBOM mutableMetadataBOM, MdsPackage mdsPackage) {
        if (mdsPackage.getVersions().get(0).getFiles() != null) {
            mdsPackage.getVersions().get(0).getFiles().addAll(mdsFilesFromBomArtifacts(mutableMetadataBOM));
        }
    }

    private void addVersionFromBom(MutableMetadataEntityBOM mutableMetadataBOM, MdsPackage mdsPackage) {
        List<MdsVersion> versions = mdsPackage.getVersions();
        if (versions != null) {
            versions.add(metadataEntityMapper.metadataBomToMdsVersion(mutableMetadataBOM));
        }
    }

    private List<MdsFile> mdsFilesFromBomArtifacts(MutableMetadataEntityBOM mutableMetadataBOM) {
        return mutableMetadataBOM.getArtifactMetadata().stream()
                .map(artifact -> metadataEntityMapper.metadataBomToMdsFile(artifact))
                .collect(Collectors.toList());
    }

    private MetadataEvent createMetadataEvent(PrioritizedNodeEvent nodeEvent, MetadataEntity metadataEntity,
            MetadataEventEntity eventEntity) {
        return MetadataEvent.builder()
                .eventType(nodeEvent.getType())
                .metadataEventEntity(eventEntity)
                .eventTime(nodeEvent.getTimestamp())
                .metadataEntity(metadataEntity)
                .path(nodeEvent.getPath())
                .build();
    }

    private Optional<MutableMetadataEntityBOM> createMutableMetadataEntityBOM(ItemInfo itemInfo,
            Properties properties, StatsInfo stats) {
        Optional<AbstractMetadataProvider> metadataProvider = metadataProviderForItem(itemInfo.getRepoKey());
        if (metadataProvider.isPresent()) {
            MutableMetadataEntityBOM metadataBom = new MutableMetadataEntityBOM();
            metadataProvider.get().supplement(metadataBom, (FileInfo) itemInfo, properties, stats);
            return Optional.of(metadataBom);
        }
        return Optional.empty();
    }

    @Nonnull
    private StatsService statsService() {
        return ContextHelper.get().beanForType(StatsService.class);
    }

    private Optional<AbstractMetadataProvider> metadataProviderForItem(String repoKey) {
        LocalRepoDescriptor repoDescriptor = getLocalOrCacheRepoDescriptor(repoKey);
        if (repoDescriptor != null) {
            return Optional.ofNullable(metadataProvidersMap.get(repoDescriptor.getType()));
        }
        log.debug("Metadata provider not found for repo '{}'", repoKey);
        return Optional.empty();
    }

    private LocalRepoDescriptor getLocalOrCacheRepoDescriptor(String repoKey) {
        LocalRepoDescriptor repoDescriptor = repositoryService.localOrCachedRepoDescriptorByKey(repoKey);
        if (repoDescriptor == null) {
            log.debug("Unable to associate Metadata event with a local or cache repository.");
            return null;
        }
        return repoDescriptor;
    }

    void setMetadataClient(MetadataClient metadataClient) {
        this.metadataClient = metadataClient;
    }

    @Autowired
    void setPropertiesService(PropertiesService propertiesService) {
        this.propertiesService = propertiesService;
    }

    @Autowired
    void setFileService(FileService fileService) {
        this.fileService = fileService;
    }

    @Autowired
    void setRepositoryService(RepositoryService repositoryService) {
        this.repositoryService = repositoryService;
    }

    @Autowired
    void setAccessService(AccessService accessService) {
        this.accessService = accessService;
    }

    @Autowired
    void setNodeEventCursorService(NodeEventCursorService nodeEventCursorService) {
        this.nodeEventCursorService = nodeEventCursorService;
    }

    @Autowired
    void setNodeEventTaskManager(NodeEventTaskManager nodeEventTaskManager) {
        this.nodeEventTaskManager = nodeEventTaskManager;
    }

    @Autowired
    void setStatusProbeService(StatusProbeService statusProbeService) {
        this.statusProbeService = statusProbeService;
    }

    @Autowired
    public void setMigrationStatusStorageService(
            MigrationStatusStorageService migrationStatusStorageService) {
        this.migrationStatusStorageService = migrationStatusStorageService;
    }

    @Autowired
    public void setNodesDao(NodesDao nodesDao) {
        this.nodesDao = nodesDao;
    }

    @Autowired
    public void setAqlService(AqlService aqlService) {
        this.aqlService = aqlService;
    }

    @Autowired
    public void setEventsService(EventsService eventsService) {
        this.eventsService = eventsService;
    }
}
