package org.artifactory.metadata.service;

import org.artifactory.aql.AqlService;
import org.artifactory.aql.api.domain.sensitive.AqlApiItem;
import org.artifactory.event.EventOperatorId;
import org.artifactory.event.provider.EventLogTableEventProvider;
import org.artifactory.event.provider.NodesTableEventProvider;
import org.artifactory.event.work.NodeEventTaskManager;
import org.artifactory.metadata.migration.MetadataMigrationInfoBlob;
import org.artifactory.storage.db.event.model.NodeEventCursor;
import org.artifactory.storage.db.event.service.NodeEventCursorService;
import org.artifactory.storage.db.fs.dao.NodesDao;
import org.artifactory.storage.db.migration.model.MigrationStatus;
import org.artifactory.storage.db.migration.service.MigrationStatusStorageService;
import org.artifactory.storage.event.EventsService;
import org.jfrog.common.ClockUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Optional;
import java.util.concurrent.Callable;
import java.util.function.Consumer;

import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.*;

/**
 * @author Uriah Levy
 */
class MetadataMigrationHelper {
    private static final Logger log = LoggerFactory.getLogger(MetadataMigrationHelper.class);
    private MigrationStatusStorageService migrationStatusStorageService;
    private NodeEventCursorService nodeEventCursorService;
    private NodesDao nodesDao;
    private AqlService aqlService;
    private EventsService eventsService;
    private long currentEventLogTimestamp;
    private long lastNodeId;

    static final String MIGRATION_IDENTIFIER = "metadata-service-migration";

    MetadataMigrationHelper(MigrationStatusStorageService migrationStatusStorageService,
            NodeEventCursorService nodeEventCursorService, NodesDao nodesDao, AqlService aqlService,
            EventsService eventsService) {
        this.migrationStatusStorageService = migrationStatusStorageService;
        this.nodeEventCursorService = nodeEventCursorService;
        this.nodesDao = nodesDao;
        this.aqlService = aqlService;
        this.eventsService = eventsService;
    }

    void migrateOrStartEventPipe(MetadataEventService metadataEventService, NodeEventTaskManager nodeEventTaskManager) {
        if (shouldMigrate()) {
            handleMigration(nodeEventTaskManager, metadataEventService);
        } else {
            nodeEventTaskManager
                    .startWithInfiniteProvider(metadataEventService, new EventLogTableEventProvider(eventsService));
        }
    }

    boolean shouldMigrate() {
        if (isNodesTableEmpty()) {
            insertNewCompletedMigration();
            return false;
        }
        Optional<MigrationStatus> migrationStatus = getMigrationStatus();
        return migrationStatus
                .map(status -> shouldMigrateByMigrationStatus(
                        nodeEventCursorService.cursorForOperator(EventOperatorId.METADATA_OPERATOR_ID.getId()), status))
                .orElse(true);
    }

    void initMigrationState() {
        MigrationStatus metadataMigration = migrationStatusStorageService
                .findMigrationByIdWithInfoBlob(MIGRATION_IDENTIFIER, MetadataMigrationInfoBlob.class);
        if (metadataMigration == null) {
            handleMissingMigration();
        } else {
            lastNodeId = getMigrationTargetNodeId();
            currentEventLogTimestamp = ((MetadataMigrationInfoBlob) metadataMigration.getMigrationInfoBlob())
                    .getCurrentEventLogTimestamp();
        }
    }

    void handleMissingMigration() {
        if (isNodesTableEmpty()) {
            // Nothing to migrate, mark done
            insertNewCompletedMigration();
            return;
        }
        currentEventLogTimestamp = eventsService.getLastEventLogTimestamp();
        insertNewDueMigration();
    }

    long getMigrationTargetNodeId() {
        MigrationStatus metadataMigration = migrationStatusStorageService
                .findMigrationByIdWithInfoBlob(MIGRATION_IDENTIFIER, MetadataMigrationInfoBlob.class);
        return ((MetadataMigrationInfoBlob) metadataMigration
                .getMigrationInfoBlob()).getMigrationTargetNodeId();
    }

    private void handleMigration(NodeEventTaskManager nodeEventTaskManager, MetadataEventService metadataEventService) {
        initMigrationState();
        nodeEventTaskManager.startWithFiniteProvider(metadataEventService,
                new NodesTableEventProvider(aqlService, onBetweenBatches(),
                        onFinish(nodeEventTaskManager, metadataEventService), this::migrationBatchQuery));
    }

    private Runnable onFinish(NodeEventTaskManager nodeEventTaskManager, MetadataEventService metadataEventService) {
        return () -> {
            migrationStatusStorageService.migrationDone(MIGRATION_IDENTIFIER);
            nodeEventTaskManager.stopDispatcherThread(metadataEventService.getOperatorId());
            nodeEventCursorService.updateCursor(
                    new NodeEventCursor(EventOperatorId.METADATA_OPERATOR_ID.getId(), currentEventLogTimestamp));
            nodeEventTaskManager
                    .startWithInfiniteProvider(metadataEventService, new EventLogTableEventProvider(eventsService));
        };
    }

    private AqlApiItem migrationBatchQuery(NodeEventCursor currentCursor) {
        return create()
                .filter(and(
                        itemId().greater(currentCursor.getEventMarker()),
                        itemId().lessEquals(lastNodeId)
                        )
                )
                .include(path(), name(), itemId())
                .limit(1000)
                .addSortElement(itemId()).asc();
    }

    private <T> T throwingCallable(Callable<T> callable) {
        try {
            return callable.call();
        } catch (Exception e) {
            throw new IllegalStateException(e.getMessage(), e);
        }
    }

    void insertNewDueMigration() {
        log.info("Migration Status is missing for '{}'. Triggering the migration process.", MIGRATION_IDENTIFIER);
        lastNodeId = findLastNodeId(); // migration target
        MigrationStatus migrationStatus = new MigrationStatus(MIGRATION_IDENTIFIER, ClockUtils.epochMillis(), 0,
                new MetadataMigrationInfoBlob(lastNodeId, currentEventLogTimestamp));
        migrationStatusStorageService.insertMigrationWithInfoBlob(migrationStatus);
    }

    void insertNewCompletedMigration() {
        log.info("Inserting completed migration status for '{}'", MIGRATION_IDENTIFIER);
        MigrationStatus migrationStatus = new MigrationStatus(MIGRATION_IDENTIFIER, ClockUtils.epochMillis(),
                ClockUtils.epochMillis(), null);
        migrationStatusStorageService.insertMigration(migrationStatus);
    }

    private boolean shouldMigrateByMigrationStatus(NodeEventCursor currentCursor, MigrationStatus migrationStatus) {
        if (migrationStatus.getFinished() != 0) { // Migration done
            return false;
        }
        MetadataMigrationInfoBlob migrationInfoBlob = (MetadataMigrationInfoBlob) migrationStatus
                .getMigrationInfoBlob();
        if (!migrationTargetReached(currentCursor, migrationInfoBlob.getMigrationTargetNodeId())) {
            log.info(
                    "Current operator cursor is at '{}', and migration target is '{}'. Starting / continuing migration.",
                    currentCursor.getEventMarker(), migrationInfoBlob.getMigrationTargetNodeId());
            return true;
        }
        log.debug("Current operator cursor is at '{}', and migration target is '{}'. Metadata Migration will not run.",
                currentCursor.getEventMarker(), migrationInfoBlob.getMigrationTargetNodeId());
        return false;
    }

    private boolean migrationTargetReached(NodeEventCursor currentCursor, long migrationTargetNodeId) {
        return currentCursor.getEventMarker() >= migrationTargetNodeId;
    }

    private Consumer<Integer> onBetweenBatches() {
        return resultSetSize -> log
                .info("Metadata Migration is handling a batch of {} results", resultSetSize);
    }

    private boolean isNodesTableEmpty() {
        if (findLastNodeId() == 0) {
            log.info("No node candidates found for migration. Migration will not run");
            return true;
        }
        return false;
    }

    private long findLastNodeId() {
        return throwingCallable(nodesDao::findLastNodeId);
    }

    private Optional<MigrationStatus> getMigrationStatus() {
        return Optional.ofNullable(migrationStatusStorageService
                .findMigrationByIdWithInfoBlob(MIGRATION_IDENTIFIER, MetadataMigrationInfoBlob.class));
    }
}
