package org.artifactory.metadata.service.provider;

import com.google.common.collect.ImmutableList;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.aql.AqlService;
import org.artifactory.aql.api.domain.sensitive.AqlApiItem;
import org.artifactory.aql.result.AqlEagerResult;
import org.artifactory.aql.result.rows.AqlItem;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.fs.FileInfo;
import org.artifactory.fs.StatsInfo;
import org.artifactory.md.Properties;
import org.artifactory.repo.RepoPath;
import org.artifactory.storage.db.event.service.metadata.model.MutableArtifactMetadata;
import org.artifactory.storage.db.event.service.metadata.model.MutableMetadataEntityBOM;
import org.jfrog.metadata.client.model.MdsVersionRepo;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;
import java.util.function.BiConsumer;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * @author Uriah Levy
 * Copies data into the {@link MutableMetadataEntityBOM}, either from the {@link Properties}, or some other data source
 */
public abstract class AbstractMetadataProvider implements MetadataProvider {

    private static final Logger log = LoggerFactory.getLogger(AbstractMetadataProvider.class);

    private RepoType type;
    private Function<RepoPath, AqlApiItem> versionFilesQuery;
    private static final String ARTIFACTORY_LICENSES = "artifactory.licenses";
    private Set<BiConsumer<MutableMetadataEntityBOM, Properties>> propertyResolvers = new HashSet<>();
    private Set<BiConsumer<MutableMetadataEntityBOM, FileInfo>> fileInfoResolvers = new HashSet<>();
    private Set<Consumer<MutableMetadataEntityBOM>> genericResolvers = new HashSet<>();
    private TypeSpecificMetadataTemplate typeSpecificMetadataTemplate;

    AbstractMetadataProvider(RepoType type, Function<RepoPath, AqlApiItem> versionFilesQuerySupplier) {
        this.type = type;
        this.versionFilesQuery = versionFilesQuerySupplier;
        initResolvers();
    }

    /**
     * Copy data from the {@link Properties} into the metadata BOM
     *
     * @param metadataBOM - the {@link MutableMetadataEntityBOM} that will hold the metadata
     * @param fileInfo    - the {@link FileInfo} object that containing file related metadata
     * @param properties  - the {@link Properties} key-value store containing package metadata
     */
    @Override
    public void supplement(MutableMetadataEntityBOM metadataBOM, FileInfo fileInfo, Properties properties,
            StatsInfo stats) {
        propertyResolvers.forEach(resolver -> resolver.accept(metadataBOM, properties)); // property metadata
        fileInfoResolvers.forEach(resolver -> resolver.accept(metadataBOM, fileInfo)); // file-info metadata
        genericResolvers.forEach(resolver -> resolver.accept(metadataBOM)); // static type-bound metadata
        statsResolver().accept(metadataBOM, stats);
    }

    @Override
    public RepoType getRepoType() {
        return type;
    }

    public abstract boolean isLeadArtifact(RepoPath repoPath);

    void setTypeSpecificMetadataTemplate(TypeSpecificMetadataTemplate typeSpecificMetadataTemplate) {
        this.typeSpecificMetadataTemplate = typeSpecificMetadataTemplate;
    }

    private void initResolvers() {
        propertyResolvers.addAll(ImmutableList.of(
                nameResolver(),
                versionResolver(),
                licenseResolver(),
                pkgIdResolver(),
                userPropsResolver(),
                qualifiersResolver()));
        fileInfoResolvers.add(fileInfoResolver());
        genericResolvers.add(typeResolver());
    }

    private BiConsumer<MutableMetadataEntityBOM, Properties> versionResolver() {
        return (MutableMetadataEntityBOM bom, Properties properties) -> {
            String version = valueForKey(properties, typeSpecificMetadataTemplate.getVersionKey());
            if (StringUtils.isNotBlank(version)) {
                bom.setVersion(version);
            }
        };
    }

    private BiConsumer<MutableMetadataEntityBOM, StatsInfo> statsResolver() {
        return (MutableMetadataEntityBOM bom, StatsInfo stats) -> {
            if (stats != null) {
                bom.setDownloadCount(stats.getDownloadCount());
            }
        };
    }

    private BiConsumer<MutableMetadataEntityBOM, Properties> userPropsResolver() {
        return (MutableMetadataEntityBOM bom, Properties properties) -> properties.keySet()
                .forEach(key -> bom.getUserProperties().put(key, properties.get(key)));
    }

    private BiConsumer<MutableMetadataEntityBOM, Properties> qualifiersResolver() {
        return (MutableMetadataEntityBOM bom, Properties properties) -> typeSpecificMetadataTemplate.getQualifierKeys()
                .keySet()
                .forEach(qualifierKey -> {
                    String rawValue = valueForKey(properties, qualifierKey);
                    if (StringUtils.isNotBlank(rawValue)) {
                        String value = valueByTemplateFunction(rawValue,
                                typeSpecificMetadataTemplate.getQualifierValueResolver());
                        if (StringUtils.isNotBlank(value)) {
                            bom.getQualifiers().put(typeSpecificMetadataTemplate.getQualifierKeys().get(qualifierKey),
                                    value);
                        }
                    }
                });
    }

    private BiConsumer<MutableMetadataEntityBOM, Properties> nameResolver() {
        return (MutableMetadataEntityBOM bom, Properties properties) -> {
            String name = valueForKey(properties, typeSpecificMetadataTemplate.getNameKey());
            if (StringUtils.isNotBlank(name)) {
                bom.setName(name);
            }
        };
    }

    private BiConsumer<MutableMetadataEntityBOM, Properties> pkgIdResolver() {
        return (MutableMetadataEntityBOM bom, Properties properties) -> {
            String name = valueForKey(properties, typeSpecificMetadataTemplate.getNameKey());
            if (StringUtils.isNotBlank(name)) {
                bom.setPkgid(getRepoType().getEffectiveType() + ":" + name);
            }
        };
    }

    private BiConsumer<MutableMetadataEntityBOM, Properties> licenseResolver() {
        return (MutableMetadataEntityBOM bom, Properties properties) -> {
            Set<String> licenses = collectionForKey(properties, ARTIFACTORY_LICENSES);
            if (CollectionUtils.isNotEmpty(licenses)) {
                bom.setLicenses(new ArrayList<>(licenses));
            }
        };
    }

    private Consumer<MutableMetadataEntityBOM> typeResolver() {
        return (MutableMetadataEntityBOM bom) -> bom.setPackageType(type.getEffectiveType());
    }

    private BiConsumer<MutableMetadataEntityBOM, FileInfo> fileInfoResolver() {
        return (MutableMetadataEntityBOM bom, FileInfo leadArtifactFileInfo) -> {
            setArtifactMetadata(bom, leadArtifactFileInfo);
            RepositoryService repositoryService = ContextHelper.get().beanForType(RepositoryService.class);
            boolean isLocal =
                    !repositoryService.localOrCachedRepoDescriptorByKey(leadArtifactFileInfo.getRepoKey()).isCache()
                            && repositoryService.localOrCachedRepoDescriptorByKey(leadArtifactFileInfo.getRepoKey())
                            .isLocal();
            setVersionRepo(bom, leadArtifactFileInfo, isLocal);
        };
    }

    private void setVersionRepo(MutableMetadataEntityBOM bom, FileInfo leadArtifactFileInfo, boolean isLocal) {
        String baseUrl = ContextHelper.get().beanForType(CentralConfigService.class).getDescriptor()
                .getUrlBase();
        String repoUrl = null;
        if (!StringUtils.isBlank(baseUrl)) {
            repoUrl = baseUrl + "/" + leadArtifactFileInfo.getRepoPath().toPath();
        } else {
            log.debug("The Artifactory Base URL is not set. Not populating the repository URL.");
        }
        MdsVersionRepo mdsVersionRepo = new MdsVersionRepo(leadArtifactFileInfo.getRepoKey(),
                repoUrl, isLocal ? "local" : "remote", leadArtifactFileInfo.getRepoPath().getPath());
        bom.setRepos(Collections.singletonList(mdsVersionRepo));
    }

    private void setArtifactMetadata(MutableMetadataEntityBOM bom, FileInfo leadArtifactFileInfo) {
        List<MutableArtifactMetadata> versionFiles = findVersionFiles(leadArtifactFileInfo);
        MutableArtifactMetadata leadArtifact = new MutableArtifactMetadata(
                leadArtifactFileInfo.getName(), leadArtifactFileInfo.getMimeType(), leadArtifactFileInfo.getSize(),
                leadArtifactFileInfo.getMd5(), leadArtifactFileInfo.getSha1(), leadArtifactFileInfo.getSha2(),
                true);
        versionFiles.add(leadArtifact);
        bom.setArtifactMetadata(versionFiles);
    }

    private List<MutableArtifactMetadata> findVersionFiles(FileInfo leadArtifactFileInfo) {
        if (versionFilesQuery != null) {
            AqlService aqlService = ContextHelper.get().beanForType(AqlService.class);
            AqlEagerResult<AqlItem> aqlResult = aqlService
                    .executeQueryEager(versionFilesQuery.apply(leadArtifactFileInfo.getRepoPath()));
            if (!aqlResult.getResults().isEmpty()) {
                return aqlResult.getResults().stream()
                        .filter(item -> !item.getName().equals(leadArtifactFileInfo.getName()))
                        .map(item -> new MutableArtifactMetadata(item.getName(), null, item.getSize(),
                                item.getActualMd5(), item.getActualSha1(), item.getSha2(), false))
                        .collect(Collectors.toList());
            }
        }
        return new ArrayList<>();
    }

    private Set<String> collectionForKey(Properties properties, String key) {
        Set<String> collection = properties.get(key);
        if (CollectionUtils.isEmpty(collection)) {
            if (log.isDebugEnabled()) {
                log.debug("Artifact Metadata provider has no metadata for property '{}'", key);
            }
            return Collections.emptySet();
        }
        return collection;
    }

    private String valueByTemplateFunction(String value, Function<String, String> function) {
        if (StringUtils.isBlank(value)) {
            if (log.isDebugEnabled()) {
                log.debug("Skipping extraction of value by function from an empty value");
            }
            return "";
        }
        return function.apply(value);
    }

    private String valueForKey(Properties properties, String key) {
        String value = properties.getFirst(key);
        if (StringUtils.isBlank(value)) {
            if (log.isDebugEnabled()) {
                log.debug("Artifact Metadata provider has no metadata for property '{}'", key);
            }
            return "";
        }
        return value;
    }
}
