/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.repo.service.trash.prune;

import org.apache.commons.lang.StringUtils;
import org.artifactory.api.common.BasicStatusHolder;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.properties.PropertiesService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.aql.AqlService;
import org.artifactory.aql.api.domain.sensitive.AqlApiItem;
import org.artifactory.aql.result.rows.AqlBaseFullRowImpl;
import org.artifactory.aql.result.rows.AqlItem;
import org.artifactory.aql.util.AqlUtils;
import org.artifactory.common.ArtifactoryHome;
import org.artifactory.common.ConstantValues;
import org.artifactory.descriptor.trashcan.TrashcanConfigDescriptor;
import org.artifactory.md.Properties;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.storage.binstore.service.GarbageCollectListenerResultStatus;
import org.artifactory.storage.binstore.service.GarbageCollectorListener;
import org.artifactory.util.TimeUnitFormat;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.*;
import static org.artifactory.aql.api.internal.AqlBase.and;

/**
 * @author Shay Yaakov
 */
public class TrashcanPruner implements GarbageCollectorListener {
    private static final Logger log = LoggerFactory.getLogger(TrashcanPruner.class);

    private final AqlService aqlService;
    private final RepositoryService repoService;
    private final PropertiesService propertiesService;
    private final CentralConfigService centralConfigService;
    private final TrashService trashService;

    public TrashcanPruner() {
        this.aqlService = ContextHelper.get().beanForType(AqlService.class);
        this.repoService = ContextHelper.get().beanForType(RepositoryService.class);
        this.propertiesService = ContextHelper.get().beanForType(PropertiesService.class);
        this.centralConfigService = ContextHelper.get().getCentralConfig();
        this.trashService = ContextHelper.get().beanForType(TrashService.class);
    }

    @Override
    public GarbageCollectListenerResultStatus start() {
        return pruneTrashItems();
    }

    @Override
    public void finished() {

    }

    @Override
    public void destroy() {

    }

    private GarbageCollectListenerResultStatus pruneTrashItems() {
        GarbageCollectListenerResultStatus status;

        // No action if retention is 0 (or less)
        TrashcanConfigDescriptor trashcanConfig = centralConfigService.getDescriptor().getTrashcanConfig();
        int retentionPeriodDays = trashService.normalizeRetentionDays(trashcanConfig.getRetentionPeriodDays());
        if (!trashcanConfig.isEnabled() || retentionPeriodDays <= 0) {
            return GarbageCollectListenerResultStatus.COMPLETED;
        }

        long start = System.currentTimeMillis();
        long validFrom = start - TimeUnit.MILLISECONDS.convert(retentionPeriodDays, TimeUnit.DAYS);
        log.debug("Starting trashcan pruning with valid from date: '{}'", validFrom);

        String batchCleanupProp = ArtifactoryHome.get().getArtifactoryProperties().getProperty(
                "artifactory.gc.batch.cleanup", "false");
        boolean batchTrashcanCleanupUsingPropsQuery = Boolean.valueOf(batchCleanupProp);
        if (batchTrashcanCleanupUsingPropsQuery) {
            status = cleanTrashcanBatch(String.valueOf(validFrom));
        } else {
            cleanTrashcan(validFrom);
            status = GarbageCollectListenerResultStatus.COMPLETED;
        }
        long end = System.currentTimeMillis();
        String duration = TimeUnitFormat.getTimeString((end - start), TimeUnit.MILLISECONDS);
        log.info("Trashcan pruning total execution time: '{}'", duration);
        return status;
    }

    private GarbageCollectListenerResultStatus cleanTrashcanBatch(String validFrom) {
        log.debug("Attempting to delete trashcan candidates");
        int undeployedItems = 0;
        try {
            undeployedItems = doCleanTrashcanBatch(validFrom);
            log.debug("{} trashcan items were deleted.", undeployedItems);
        } catch (Exception e) {
            log.warn("Error occurred while during trashcan cleanup batch, skipping batch. {}", e.getMessage());
            log.debug("Error occurred while during trashcan cleanup batch, skipping batch.", e);
        }
        if (undeployedItems > 0) {
            // artifacts were deleted, meaning that there might be more artifacts to delete.
            return GarbageCollectListenerResultStatus.HAS_MORE_WORK;
        }
        return GarbageCollectListenerResultStatus.COMPLETED;
    }

    private void cleanTrashcan(long validFrom) {
        int maxSearchResult = ConstantValues.trashcanMaxSearchResults.getInt();
        List<AqlItem> results;
        long numberOfDeletedPackages;
        int offset = 0;
        do {
            results = findTrashItems(offset, maxSearchResult);
            numberOfDeletedPackages = results.stream()
                    .map(item -> deleteItemIfOld(item, validFrom))
                    .filter(Objects::nonNull)
                    .count();
            offset += maxSearchResult - numberOfDeletedPackages;
        } while (results.size() == maxSearchResult);
    }

    public int doCleanTrashcanBatch(String validFrom) {
        long start = System.currentTimeMillis();
        List<RepoPath> itemsToDelete = trashService.getTrashcanItemsBatchToDelete(validFrom);
        log.debug("Fetching trashcan delete candidates batch took: {} millis", System.currentTimeMillis() - start);
        try {
            itemsToDelete.stream()
                    .distinct()
                    .forEach(repoPath -> {
                        long itemDeleteStartTime = System.currentTimeMillis();
                        log.debug("Deleting item: {}", repoPath);
                        repoService.undeploy(repoPath, false, false);
                        log.debug("Deleting item: {} took {} millis", repoPath,
                                System.currentTimeMillis() - itemDeleteStartTime);
                    });
        } catch (Exception e) {
            log.warn("Error occurred while undeploying trashcan item. {}", e.getMessage());
            log.debug("Error occurred while undeploying trashcan item. ", e);
        }
        return itemsToDelete.size();
    }

    private List<AqlItem> findTrashItems(int offset, int maxSearchResult) {
        AqlApiItem query = AqlApiItem.createWithEmptyResults()
                .filter(
                        and(
                                repo().equal(TrashService.TRASH_KEY),
                                AqlApiItem.property().key().equal(TrashService.PROP_TRASH_TIME))
                )
                .include(repo(), path(), name())
                .offset(offset)
                .addSortElement(itemId()).asc()
                .limit(maxSearchResult);
        return aqlService.executeQueryEager(query).getResults();
    }

    private BasicStatusHolder deleteItemIfOld(AqlItem item, long validFrom) {
        RepoPath repoPath = AqlUtils.fromAql((AqlBaseFullRowImpl) item);
        Properties itemProperties = propertiesService.getProperties(repoPath);
        if (itemProperties.isEmpty()) {
            log.debug("Unable to find properties for '{}'", repoPath);
            return null;
        }

        String trashTimeProperty = itemProperties.getFirst(TrashService.PROP_TRASH_TIME);
        if (StringUtils.isBlank(trashTimeProperty)) {
            log.warn("Unable to find trash.time property for '{}'", repoPath);
            return null;
        }

        try {
            long trashTime = Long.parseLong(trashTimeProperty);
            if (trashTime < validFrom) {
                log.debug("Removing old trash item '{}' with trash time of '{}'", repoPath, trashTime);
                return repoService.undeploy(repoPath, false, false);
            }
        } catch (Exception e) {
            log.error("Error during deletion of the old trash item '{}': '{}'", repoPath, e.getMessage());
            log.debug("Error during deletion old trash item: " + e.getMessage(), e);
        }
        return null;
    }
}
