package org.artifactory.metadata.service;

import org.artifactory.aql.AqlService;
import org.artifactory.event.EventOperatorId;
import org.artifactory.metadata.migration.MetadataMigrationInfoBlob;
import org.artifactory.storage.db.event.model.NodeEventCursor;
import org.artifactory.storage.db.event.service.NodeEventCursorService;
import org.artifactory.storage.db.fs.dao.NodesDao;
import org.artifactory.storage.db.migration.model.MigrationStatus;
import org.artifactory.storage.db.migration.service.MigrationStatusStorageService;
import org.artifactory.storage.event.EventsService;
import org.jfrog.common.ClockUtils;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.sql.SQLException;

import static org.artifactory.metadata.service.MetadataMigrationHelper.MIGRATION_IDENTIFIER;
import static org.mockito.Mockito.*;
import static org.testng.Assert.assertTrue;
import static org.testng.AssertJUnit.assertFalse;

/**
 * @author Uriah Levy
 */
@Test
public class MetadataMigrationHelperTest {

    @Mock
    MigrationStatusStorageService migrationStatusStorageService;

    @Mock
    NodeEventCursorService cursorService;

    @Mock
    NodesDao nodesDao;

    @Mock
    AqlService aqlService;

    @Mock
    EventsService eventsService;

    @BeforeMethod
    private void beforeMethod() {
        MockitoAnnotations.initMocks(this);
    }

    @Test(expectedExceptions = IllegalStateException.class)
    public void shouldMigrateThrowsRuntimeAfterSqlException() throws SQLException {
        MetadataMigrationHelper metadataMigrationHelper = new MetadataMigrationHelper(migrationStatusStorageService,
                cursorService, nodesDao, aqlService, eventsService);
        when(nodesDao.findLastNodeId()).thenThrow(SQLException.class);
        metadataMigrationHelper.shouldMigrate();
    }

    public void shouldMigrateFalseOnNoNodes() throws SQLException {
        MetadataMigrationHelper metadataMigrationHelper = new MetadataMigrationHelper(migrationStatusStorageService,
                cursorService, nodesDao, aqlService, eventsService);

        when(nodesDao.findLastNodeId()).thenReturn(0L);

        assertFalse(metadataMigrationHelper.shouldMigrate());
        verify(migrationStatusStorageService).insertMigration(any());
    }

    public void testHandleMissingMigration() throws SQLException {
        MetadataMigrationHelper metadataMigrationHelper = new MetadataMigrationHelper(migrationStatusStorageService,
                cursorService, nodesDao, aqlService, eventsService);

        long targetNodeId = 11245;
        when(nodesDao.findLastNodeId()).thenReturn(targetNodeId);
        long lastEventLogTimestamp = ClockUtils.epochMillis();
        when(eventsService.getLastEventLogTimestamp()).thenReturn(lastEventLogTimestamp);

        metadataMigrationHelper.handleMissingMigration();
        verify(migrationStatusStorageService)
                .insertMigrationWithInfoBlob(argThat(migrationStatus ->
                        ((MetadataMigrationInfoBlob) migrationStatus.getMigrationInfoBlob())
                                .getMigrationTargetNodeId() == targetNodeId &&
                                ((MetadataMigrationInfoBlob) migrationStatus.getMigrationInfoBlob())
                                        .getCurrentEventLogTimestamp() == lastEventLogTimestamp));

    }

    public void testHandleMissingMigrationEmptyNodesTable() throws SQLException {
        MetadataMigrationHelper metadataMigrationHelper = new MetadataMigrationHelper(migrationStatusStorageService,
                cursorService, nodesDao, aqlService, eventsService);

        when(nodesDao.findLastNodeId()).thenReturn(0L);

        metadataMigrationHelper.handleMissingMigration();
        verify(migrationStatusStorageService).insertMigration(argThat(a -> a.getMigrationInfoBlob() == null));
    }

    public void testShouldMigrateByCursorTrue() throws SQLException {
        MetadataMigrationHelper metadataMigrationHelper = new MetadataMigrationHelper(migrationStatusStorageService,
                cursorService, nodesDao, aqlService, eventsService);

        when(nodesDao.findLastNodeId()).thenReturn(ClockUtils.epochMillis());
        NodeEventCursor currentCursor = new NodeEventCursor(EventOperatorId.METADATA_OPERATOR_ID.getId(),
                ClockUtils.epochMillis() + 100);

        when(cursorService.cursorForOperator(EventOperatorId.METADATA_OPERATOR_ID.getId())).thenReturn(currentCursor);
        when(migrationStatusStorageService.findMigrationByIdWithInfoBlob(MIGRATION_IDENTIFIER,
                MetadataMigrationInfoBlob.class))
                .thenReturn(new MigrationStatus(MIGRATION_IDENTIFIER, ClockUtils.epochMillis()
                        - 10_000, 0, new MetadataMigrationInfoBlob(ClockUtils.epochMillis() + Integer.MAX_VALUE, 0)));
        assertTrue(metadataMigrationHelper.shouldMigrate());
    }

    public void testInitMigrationStatusWithExistingMigration() {
        MetadataMigrationHelper metadataMigrationHelper = spy(new MetadataMigrationHelper(migrationStatusStorageService,
                cursorService, nodesDao, aqlService, eventsService));

        when(migrationStatusStorageService.findMigrationById(MIGRATION_IDENTIFIER))
                .thenReturn(new MigrationStatus(MIGRATION_IDENTIFIER, ClockUtils.epochMillis()
                        - 10_000, 0, new MetadataMigrationInfoBlob(ClockUtils.epochMillis() + Integer.MAX_VALUE, 0)));
        when(migrationStatusStorageService
                .findMigrationByIdWithInfoBlob(MIGRATION_IDENTIFIER, MetadataMigrationInfoBlob.class))
                .thenReturn(new MigrationStatus(MIGRATION_IDENTIFIER, ClockUtils.epochMillis()
                        - 10_000, 0, new MetadataMigrationInfoBlob(ClockUtils.epochMillis() + Integer.MAX_VALUE, 0)));

        metadataMigrationHelper.initMigrationState();

        verify(eventsService, times(0)).getLastEventLogTimestamp();
        verify(metadataMigrationHelper).getMigrationTargetNodeId();
    }

    public void testInitMigrationStatusWithNonExistingMigration() throws SQLException {
        MetadataMigrationHelper metadataMigrationHelper = spy(new MetadataMigrationHelper(migrationStatusStorageService,
                cursorService, nodesDao, aqlService, eventsService));

        when(migrationStatusStorageService.findMigrationById(MIGRATION_IDENTIFIER))
                .thenReturn(null);
        when(nodesDao.findLastNodeId()).thenReturn(10_000L);
        doNothing().when(metadataMigrationHelper).insertNewDueMigration();

        metadataMigrationHelper.initMigrationState();

        verify(metadataMigrationHelper, times(0)).insertNewCompletedMigration();
        verify(metadataMigrationHelper, times(0)).getMigrationTargetNodeId();
        verify(metadataMigrationHelper).insertNewDueMigration();
        verify(eventsService).getLastEventLogTimestamp();
    }
}