package org.artifactory.storage.db.binstore.service.garbage;

import com.google.common.util.concurrent.ThreadFactoryBuilder;
import org.artifactory.api.security.SecurityService;
import org.artifactory.common.ConstantValues;
import org.artifactory.storage.GCCandidate;
import org.artifactory.storage.binstore.service.GCProvider;
import org.artifactory.storage.binstore.service.GarbageCollectorInfo;
import org.artifactory.storage.binstore.service.GarbageCollectorStrategy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.*;

/**
 * @author Uriah Levy
 */
public class GCCandidateProducer implements Runnable {
    private static final Logger log = LoggerFactory.getLogger(GCCandidateProducer.class);
    private List<GCProvider> gcProviders;
    private ExecutorService executorService;
    ExecutorService producerExecutor;
    private volatile boolean alive = false;
    private SecurityService securityService;
    private GarbageCollectorStrategy garbageCollectorStrategy;

    public GCCandidateProducer(List<GCProvider> gcProviders, GarbageCollectorStrategy strategy,
            SecurityService securityService) {
        this.producerExecutor = new ThreadPoolExecutor(1, 1, 5L, TimeUnit.MINUTES,
                new LinkedBlockingQueue<>(1), new ThreadFactoryBuilder().setDaemon(true)
                .setNameFormat(strategy.name().toLowerCase() + "-gc-producer").build());
        this.gcProviders = gcProviders;
        this.securityService = securityService;
        this.garbageCollectorStrategy = strategy;
        int numberOfThreads = ConstantValues.gcNumberOfWorkersThreads.getInt();
        executorService = new ThreadPoolExecutor(numberOfThreads, numberOfThreads, 5L, TimeUnit.MINUTES,
                new LinkedBlockingQueue<>(ConstantValues.trashcanMaxSearchResults.getInt() * 2),
                new ThreadFactoryBuilder().setDaemon(true)
                        .setNameFormat(strategy.name().toLowerCase() + "-gc-worker-%s").build());
    }

    @Override
    public void run() {
        alive = true;
        try {
            while (alive) {
                for (GCProvider gcProvider : gcProviders) {
                    gcProvider.resetProvider();
                    log.info("Starting GC strategy '{}' with Provider '{}'", garbageCollectorStrategy,
                            gcProvider.getClass().getSimpleName());
                    List<GCCandidate> batch;
                    while (!((batch = gcProvider.getBatch()).isEmpty())) {
                        if (alive) {
                            doWork(gcProvider, batch);
                        } else {
                            log.debug("GC Producer ('{}' provider) stopped", gcProvider.getClass().getSimpleName());
                            return;
                        }
                    }
                    log.info("GC Strategy '{}' with Provider '{}' done (empty results)", garbageCollectorStrategy,
                            gcProvider.getClass().getSimpleName());
                }
                alive = false;
                log.info("Finished GC Strategy '{}'", garbageCollectorStrategy);
            }
        } finally {
            doStopProducer();
        }
    }

    private void doStopProducer() {
        alive = false;
        // always reset the provider, to close the underlying streaming provider
        gcProviders.forEach(GCProvider::resetProvider);
    }

    public void startProducer() {
        try {
            log.info("Starting Strategy '{}' with GC producer(s):", garbageCollectorStrategy);
            gcProviders.forEach(gcProvider -> log.info(gcProvider.getClass().getSimpleName()));
            producerExecutor.submit(this);
        } catch (RejectedExecutionException e) {
            log.info("Unable to schedule GC Producer thread with strategy '{}'. Producer status is alive: '{}'",
                    garbageCollectorStrategy, alive, e);
        }
    }

    public void askToStopProducer() {
        log.info("Requesting stopping of Strategy '{}' with GC producer(s):", garbageCollectorStrategy);
        gcProviders.forEach(gcProvider -> log.info(gcProvider.getClass().getSimpleName()));
        alive = false;
    }

    public boolean isProducerRunning() {
        return alive;
    }

    private void doWork(GCProvider gcProvider, List<GCCandidate> batch) {
        String providerName = gcProvider.getClass().getName();
        log.debug("Starting to work on GC batch for provider '{}'", providerName);
        List<Future<?>> futures = new ArrayList<>();
        GarbageCollectorInfo result = submitWorkers(gcProvider, batch, futures);
        waitForWorkers(gcProvider, futures, result);
        log.debug("Finished working on GC batch for provider '{}'", providerName);
    }

    private GarbageCollectorInfo submitWorkers(GCProvider gcProvider, List<GCCandidate> batch,
            List<Future<?>> futures) {
        GarbageCollectorInfo result = new GarbageCollectorInfo();
        result.candidatesForDeletion = batch.size();
        result.initialCount = -1;
        batch.forEach(
                candidate -> futures
                        .add(executorService.submit(() -> executeTask(gcProvider, candidate, result))));
        return result;
    }

    private void waitForWorkers(GCProvider gcProvider, List<Future<?>> futures,
            GarbageCollectorInfo result) {
        log.debug("Waiting for GC workers to finish batch execution");
        futures.forEach(future -> {
            try {
                future.get();
            } catch (InterruptedException | ExecutionException e) {
                log.error("Waiting interrupted", e);
                Thread.currentThread().interrupt();
            }
        });
        log.debug("Workers finished one batch.");
        result.gcEndTime = System.currentTimeMillis();
        if (gcProvider.reportAfterBatch()) {
            result.printCollectionInfo(-1);
        }
    }

    private void executeTask(GCProvider gcProvider, GCCandidate gcCandidate, GarbageCollectorInfo result) {
        SecurityContext securityContext = SecurityContextHolder.getContext();
        Authentication originalAuthentication = securityContext.getAuthentication();
        try {
            securityService.authenticateAsSystem();
            try {
                log.debug("Starting GC action on '{}'", gcCandidate);
                gcProvider.getAction().accept(gcCandidate, result);
            } catch (Exception e) {
                log.debug("Unable to execute GC action on GC candidate '{}'", gcCandidate, e);
            }
        } finally {
            securityContext.setAuthentication(originalAuthentication);
        }
    }
}
