/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.security.service;

import com.google.common.collect.Sets;
import org.artifactory.factory.InfoFactoryHolder;
import org.artifactory.model.xstream.security.AceImpl;
import org.artifactory.model.xstream.security.MutableRepoAclImpl;
import org.artifactory.model.xstream.security.RepoPermissionTargetImpl;
import org.artifactory.security.AceInfo;
import org.artifactory.security.Acl;
import org.artifactory.security.ArtifactoryPermission;
import org.artifactory.security.RepoPermissionTarget;
import org.artifactory.storage.security.service.AclStoreService;
import org.easymock.EasyMock;
import org.jfrog.common.ClockUtils;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;
import org.testng.collections.Lists;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Supplier;

import static org.testng.Assert.*;

/**
 * UnitTest of AclCacheLoader, especially the call method
 * mock the dao to create AclCache
 *
 * @author nadavy
 */
@Test
public class AclCacheLoaderTest {

    private static final String USERNAME1 = "user1";
    private static final String USERNAME2 = "user2";
    private static final String USERNAME3 = "user3";
    private static final String GROUPNAME1 = "group1";
    private static final String GROUPNAME2 = "group2";

    private static final String REPO1 = "repo1";
    private static final String REPO2 = "repo2";

    private AclCacheLoader.AclCacheItem aclCacheItem;

    private final Supplier<Collection<Acl<RepoPermissionTarget>>> getRepoAcls = this::getDownstreamAllRepoAcls;
    private final Function<RepoPermissionTarget, List<String>> getRepoKeysFunction = this::getRepoKeys;

    private final Function<Acl<RepoPermissionTarget>, Acl<RepoPermissionTarget>> createRepoAclFunction =
            (Acl<RepoPermissionTarget> repoAcl) -> InfoFactoryHolder.get()
                    .createRepoAcl(repoAcl.getPermissionTarget(), repoAcl.getAces(),
                            repoAcl.getUpdatedBy(), repoAcl.getLastUpdated(), repoAcl.getAccessIdentifier());

    /**
     * Create DAO mocks, populate new AclCache and call AclCache.
     */
    @BeforeClass
    public void populateAclInfo() {
        AclStoreService aclStoreService = EasyMock.createMock(AclStoreService.class);

        EasyMock.replay(aclStoreService);

        AclCacheLoader cacheLoader = new AclCacheLoader(getRepoAcls, getRepoKeysFunction);
        aclCacheItem = cacheLoader.call();
    }

    /**
     * Assert the different AclCacheLoader caches- groups and users
     */
    public void testAclCacheLoader() {
        Map<String, Map<String, Set<MutableRepoAclImpl>>> groupResultMap = aclCacheItem.getGroupResultMap();
        assertGroupMap(groupResultMap);

        Map<String, Map<String, Set<MutableRepoAclImpl>>> userResultMap = aclCacheItem.getUserResultMap();
        assertUserMap(userResultMap);

    }

    private List<String> getRepoKeys(RepoPermissionTarget repoPermissionTarget) {
        return repoPermissionTarget.getRepoKeys();
    }

    private Collection<Acl<RepoPermissionTarget>> getDownstreamAllRepoAcls() {
        Collection<Acl<RepoPermissionTarget>> aclInfos = Lists.newArrayList();
        aclInfos.add(getAnyAcl());
        aclInfos.add(getRepo1Acl());
        return aclInfos;
    }


    /**
     * create readers ACL, with user1 and user2.
     */
    private MutableRepoAclImpl getAnyAcl() {
        AceImpl user1 = new AceImpl(USERNAME1, false, 0);
        AceImpl user2 = new AceImpl(USERNAME2, false, 0);
        Set<AceInfo> aces = Sets.newHashSet(user1, user2);

        RepoPermissionTargetImpl pt = new RepoPermissionTargetImpl("readerT",
                Lists.newArrayList(RepoPermissionTargetImpl.ANY_LOCAL_REPO),
                Lists.newArrayList("**"),
                Lists.newArrayList(""));

        return new MutableRepoAclImpl(pt, aces, "me", ClockUtils.epochMillis());
    }

    /**
     * create deployers ACL, with user2, user3 and group1
     */
    private MutableRepoAclImpl getRepo1Acl() {
        AceImpl user2 = new AceImpl(USERNAME2, false, ArtifactoryPermission.DEPLOY.getMask());
        AceImpl user3 = new AceImpl(USERNAME3, false, ArtifactoryPermission.DEPLOY.getMask());
        AceImpl group = new AceImpl(GROUPNAME1, true, ArtifactoryPermission.DEPLOY.getMask());
        Set<AceInfo> aces = Sets.newHashSet(user2, user3, group);

        RepoPermissionTarget pt = new RepoPermissionTargetImpl("deployT", Lists.newArrayList(REPO1, REPO2),
                Lists.newArrayList("**"), Lists.newArrayList("a/**"));

        return new MutableRepoAclImpl(pt, aces, "me", ClockUtils.epochMillis());
    }

    /**
     * assert that user1 and user2 - read permission on any local repo
     * user2 and user3 2 - deploy permissions on repo1 (except a/**)
     */
    private void assertUserMap(Map<String, Map<String, Set<MutableRepoAclImpl>>> userResultMap) {
        assertEquals(3, userResultMap.size(), "UserAclMap should have 3 users");

        Map<String, Set<MutableRepoAclImpl>> user1RepoToAclMap = userResultMap.get(USERNAME1);
        Map<String, Set<MutableRepoAclImpl>> user2RepoToAclMap = userResultMap.get(USERNAME2);
        Map<String, Set<MutableRepoAclImpl>> user3RepoToAclMap = userResultMap.get(USERNAME3);

        assertEquals(1, user1RepoToAclMap.size(), "User1 should have permission on ANY");
        assertEquals(3, user2RepoToAclMap.size(), "User2 should have ANY, REPO1 and REPO2");
        assertEquals(2, user3RepoToAclMap.size(), "User3 should have REPO1 and REPO2");

        assertNotNull(user1RepoToAclMap.get(RepoPermissionTargetImpl.ANY_LOCAL_REPO),
                "User1 should have permission on ANY");
        assertNotNull(user2RepoToAclMap.get(RepoPermissionTargetImpl.ANY_LOCAL_REPO),
                "User1 should have permission on ANY");
        assertNotNull(user2RepoToAclMap.get(REPO1), "User2 should have permission on REPO1");
        assertNotNull(user2RepoToAclMap.get(REPO2), "User2 should have permission on REPO2");
        assertNotNull(user3RepoToAclMap.get(REPO1), "User3 should have permission on REPO1");
        assertNotNull(user3RepoToAclMap.get(REPO2), "User3 should have permission on REPO2");

        // check include/exclude
        assertEquals("a/**", user3RepoToAclMap.get(REPO1).iterator().next().getPermissionTarget().getExcludesPattern());
        assertEquals("**", user3RepoToAclMap.get(REPO1).iterator().next().getPermissionTarget().getIncludesPattern());
    }

    /**
     * assert that group1 is in 1 acl, has deploy permission on repo1 only
     * group2 should have any acls
     */
    private void assertGroupMap(Map<String, Map<String, Set<MutableRepoAclImpl>>> groupRepoToAclMap) {
        assertEquals(1, groupRepoToAclMap.size());
        Map<String, Set<MutableRepoAclImpl>> groupMapToAcl = groupRepoToAclMap.get(GROUPNAME1);
        Set<MutableRepoAclImpl> groupRepoAcls = groupMapToAcl
                .get(REPO1); // group1 has 1 acl, which consists only of REPO1
        assertNotNull(groupRepoAcls, "GROUP1 shouldn't be null");
        assertEquals(1, groupRepoAcls.size(), "GROUP1 should have only 1 ACL");
        Set<AceInfo> groupAces = groupRepoAcls.iterator().next().getAces();
        assertEquals(groupAces.size(), 3, "GROUP1 ACL should contain 3 ACEs"); // 2nd with user3
        assertNull(groupRepoToAclMap.get(GROUPNAME2), "GROUP2 don't have ACLs ");
    }
}
