package org.artifactory.rest.common.service.security.permissions;

import org.artifactory.rest.exception.BadRequestException;
import org.artifactory.rest.exception.ConflictException;
import org.artifactory.rest.exception.ForbiddenWebAppException;
import org.artifactory.rest.exception.NotFoundException;
import org.artifactory.security.PermissionTargetNaming;
import org.artifactory.security.SecurityEntityListItem;
import org.artifactory.security.permissions.SecurityEntityPermissionTargetModel;
import org.artifactory.security.permissions.PermissionTargetModel;

import javax.servlet.http.HttpServletRequest;
import java.util.List;
import java.util.Set;

/**
 * @author Dan Feldman
 */
public interface RestSecurityRequestHandlerV2 {

    /**
     * @return true if a permission target (build, repo or both) with a name {@param entityKey} exists
     */
    boolean isPermissionTargetExists(String entityKey) throws ForbiddenWebAppException, NotFoundException;

    /**
     * @param outputMode Governs the way actions strings are written
     * @return all permission targets for the current logged-in user for the type of permission given.
     */
    List<PermissionTargetModel> getAllPermissionTargets(PermissionTargetNaming outputMode);

    /**
     * @param request to be used to get the context url of the request
     * @return all permission targets for the current logged-in user for the type of permission given,
     * in a lightweight type of list. (name + url)
     */
    Set<SecurityEntityListItem> getAllPermissionTargets(HttpServletRequest request);

    /**
     * @param outputMode Governs the way actions strings are written
     * @return a {@link PermissionTargetModel} representing the requested {@param entityKey},
     * null if no permission targets (build or repo) were found.
     */
    PermissionTargetModel getPermissionTarget(String entityKey, PermissionTargetNaming outputMode)
            throws ForbiddenWebAppException, NotFoundException;

    /**
     * Creates permission targets.
     *
     * @param inputMode Governs the way actions strings are read
     * @throws BadRequestException On any validation error.
     * @throws ConflictException   On mismatch in permission target names.
     */
    void createPermissionTarget(String entityKey, PermissionTargetModel permissionTargetModel,
            PermissionTargetNaming inputMode) throws ConflictException, BadRequestException;

    /**
     * Updates permission targets.
     *
     * @param inputMode Governs the way actions strings are read
     * @throws NotFoundException   If the permission sent is missing (or it is hidden due to security config).
     * @throws ForbiddenWebAppException  If user has no manage (or admin) permission on requested permission
     * @throws BadRequestException On any validation error.
     */
    void updatePermissionTarget(String entityKey, PermissionTargetModel permissionTargetModel,
            PermissionTargetNaming inputMode) throws ForbiddenWebAppException, NotFoundException, BadRequestException;

    /**
     * Deletes the permission target identified by {@param entityKey}
     *
     * @throws NotFoundException   If no permission with the name given in {@param entityKey} exists.
     * @throws BadRequestException On any validation error.
     */
    void deletePermissionTarget(String entityKey) throws NotFoundException, BadRequestException;

    /**
     * Provides a list of builds that match the patterns provided.
     * Empty list if no builds were found to match.
     */
    List<String> getBuildsPerPatterns(List<String> includePatterns, List<String> excludePatterns) throws NotFoundException;

    /**
     * Retrieve A User Permission targets.
     * @param entityKey The name of the User
     */
    List<SecurityEntityPermissionTargetModel> getUserPermissionSecurityEntity(String entityKey);

    /**
     * Retrieve A Group Permission targets.
     * @param entityKey The name of the Group
     */
    List<SecurityEntityPermissionTargetModel> getGroupPermissionSecurityEntity(String entityKey);

}
