/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes;

import com.google.common.collect.Lists;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.CoreAddons;
import org.artifactory.addon.support.SupportAddon;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.repo.*;
import org.artifactory.descriptor.repo.distribution.DistributionRepoDescriptor;
import org.artifactory.descriptor.repo.releasebundle.ReleaseBundlesRepoDescriptor;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.rest.common.model.RestModel;
import org.artifactory.security.ArtifactoryPermission;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.repo.RepositoryNode;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.repo.VirtualRemoteRepositoryNode;
import org.codehaus.jackson.annotate.JsonTypeName;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;

import static org.artifactory.repo.RepoDetailsType.*;
import static org.artifactory.repo.RepoPath.REMOTE_CACHE_SUFFIX;

/**
 * This is the root node of the tree browser. It contains all the repository nodes.
 *
 * @author Chen Keinan
 */
@JsonTypeName("root")
class RootNode implements RestTreeNode {

    @Override
    public Collection<? extends RestModel> fetchItemTypeData(boolean isCompact) {
        List<RestModel> repoNodes = new ArrayList<>();
        //Add a tree node for each file repository and local cache repository
        RepositoryService repositoryService = ContextHelper.get().getRepositoryService();
        addDistributionRepoNodes(repoNodes, repositoryService);
        addLocalRepoNodes(repoNodes, repositoryService);
        addRemoteRepoNodes(repoNodes, repositoryService);
        addVirtualRepoNodes(repoNodes, repositoryService);
        addReleaseBundlesRepoNodes(repoNodes, repositoryService);
        addTrashRepo(repoNodes, repositoryService);
        addSupportBundles(repoNodes, repositoryService);
        return repoNodes;
    }

    private void addDistributionRepoNodes(List<RestModel> repoNodes, RepositoryService repositoryService) {
        List<DistributionRepoDescriptor> distributionDescriptors = repositoryService.getDistributionRepoDescriptors();
        removeNonPermissionRepositories(distributionDescriptors);
        distributionDescriptors.sort(new RepoComparator());
        repoNodes.addAll(getDistributionNodes(distributionDescriptors));
    }

    private void addLocalRepoNodes(List<RestModel> repoNodes, RepositoryService repositoryService) {
        List<LocalRepoDescriptor> localRepos = repositoryService.getLocalAndCachedRepoDescriptors();
        removeNonPermissionRepositories(localRepos);
        localRepos.sort(new LocalRepoAlphaComparator());
        repoNodes.addAll(getLocalNodes(localRepos));
    }

    private void addRemoteRepoNodes(List<RestModel> repoNodes, RepositoryService repositoryService) {
        List<RemoteRepoDescriptor> remoteDescriptors = repositoryService.getRemoteRepoDescriptors();
        removeNonPermissionRepositories(remoteDescriptors);
        remoteDescriptors.sort(new RepoComparator());
        repoNodes.addAll(getRemoteNodes(remoteDescriptors));
    }

    private void addReleaseBundlesRepoNodes(List<RestModel> repoNodes, RepositoryService repositoryService) {
        AuthorizationService authorizationService = ContextHelper.get().getAuthorizationService();
        if (authorizationService.isAdmin()) {
            List<ReleaseBundlesRepoDescriptor> bundlesRepoDescriptors = repositoryService
                    .getReleaseBundlesRepoDescriptors();
            bundlesRepoDescriptors.sort(new RepoComparator());
            repoNodes.addAll(getReleaseBundlesNodes(bundlesRepoDescriptors));
        }
    }

    /**
     * add virtual repo nodes to repo list
     *
     * @param repoNodes         - repository nodes list
     * @param repositoryService - repository service
     */
    private void addVirtualRepoNodes(List<RestModel> repoNodes, RepositoryService repositoryService) {
        List<VirtualRepoDescriptor> virtualDescriptors = repositoryService.getVirtualRepoDescriptors();
        removeNonPermissionRepositories(virtualDescriptors);
        virtualDescriptors.sort(new RepoComparator());
        repoNodes.addAll(getVirtualNodes(virtualDescriptors));
    }

    private void addTrashRepo(List<RestModel> repoNodes, RepositoryService repositoryService) {
        AuthorizationService authService = ContextHelper.get().beanForType(AuthorizationService.class);
        if (authService.isAdmin()) {
            LocalRepoDescriptor trashDescriptor = repositoryService.localRepoDescriptorByKey(TrashService.TRASH_KEY);
            repoNodes.add(new RepositoryNode(trashDescriptor.getKey(), trashDescriptor.getType(), "trash"));
        }
    }

    private void addSupportBundles(List<RestModel> repoNodes, RepositoryService repositoryService) {
        AuthorizationService authService = ContextHelper.get().beanForType(AuthorizationService.class);
        AddonsManager addonsManager = ContextHelper.get().beanForType(AddonsManager.class);
        boolean isAol = addonsManager.addonByType(CoreAddons.class).isAol();
        boolean isSupportAddonEnabled = addonsManager.addonByType(SupportAddon.class).isSupportAddonEnabled();
        if (authService.isAdmin() && isSupportAddonEnabled && !isAol) {
            LocalRepoDescriptor supportBundles = repositoryService.localRepoDescriptorByKey(SupportBundleRepoDescriptor.SUPPORT_BUNDLE_REPO_NAME);
            repoNodes.add(new RepositoryNode(supportBundles.getKey(), supportBundles.getType(), "supportBundles"));
        }
    }

    private List<INode> getLocalNodes(List<LocalRepoDescriptor> repos) {
        List<INode> items = Lists.newArrayListWithCapacity(repos.size());
        repos.forEach(repo -> {
            String repoType = repo.getKey().endsWith(REMOTE_CACHE_SUFFIX) ? "cached" : LOCAL.typeNameLowercase();
            RepositoryNode itemNodes = new RepositoryNode(repo.getKey(), repo.getType(), repoType);
            items.add(itemNodes);
        });
        return items;
    }

    private List<INode> getRemoteNodes(List<RemoteRepoDescriptor> repos) {
        List<INode> items = Lists.newArrayListWithCapacity(repos.size());
        repos.forEach(repo -> {
            if (repo.isListRemoteFolderItems()) {
                VirtualRemoteRepositoryNode itemNodes = new VirtualRemoteRepositoryNode(repo.getKey(), repo.getType(),
                        REMOTE.typeNameLowercase(), false);
                items.add(itemNodes);
            }
        });
        return items;
    }

    private List<INode> getVirtualNodes(List<VirtualRepoDescriptor> repos) {
        List<INode> items = Lists.newArrayListWithCapacity(repos.size());
        repos.forEach(repo -> {
            VirtualRemoteRepositoryNode itemNodes = new VirtualRemoteRepositoryNode(repo.getKey(), repo.getType(),
                    VIRTUAL.typeNameLowercase(), repo.getDefaultDeploymentRepo() != null);
            items.add(itemNodes);
        });
        return items;
    }

    private List<INode> getDistributionNodes(List<DistributionRepoDescriptor> repos) {
        List<INode> items = Lists.newArrayListWithCapacity(repos.size());
        repos.forEach(repo -> items.add(new RepositoryNode(repo.getKey(), repo.getType(), DISTRIBUTION.typeNameLowercase())));
        return items;
    }

    private List<INode> getReleaseBundlesNodes(List<ReleaseBundlesRepoDescriptor> repos) {
        List<INode> items = Lists.newArrayListWithCapacity(repos.size());
        repos.forEach(repo -> items.add(new RepositoryNode(repo.getKey(), repo.getType(), "releaseBundles")));
        return items;
    }

    private void removeNonPermissionRepositories(List<? extends RepoDescriptor> repositories) {
        AuthorizationService authorizationService = ContextHelper.get().getAuthorizationService();
        repositories.removeIf(repoDescriptor ->
                !userCanReadAnywhereOnBuildRepo(authorizationService, repoDescriptor)
                        && !authorizationService.userHasPermissionsOnRepositoryRoot(repoDescriptor.getKey()));
    }

    /**
     * Special case for the build info repo, since its improbable for admins to give users permissions on its root.
     * We allow any user that has read anywhere on the repo to see it, given that individual folders and files are
     * still filtered by permissions.
     */
    private boolean userCanReadAnywhereOnBuildRepo(AuthorizationService authService, RepoDescriptor repository) {
        return RepoType.BuildInfo.equals(repository.getType()) && authService.hasBuildPermission(ArtifactoryPermission.READ);
    }

    private static class RepoComparator implements Comparator<RepoBaseDescriptor> {

        @Override
        public int compare(RepoBaseDescriptor descriptor1, RepoBaseDescriptor descriptor2) {

            //Local repositories can be either ordinary or caches
            if (descriptor1 instanceof LocalRepoDescriptor) {
                boolean repo1IsCache = ((LocalRepoDescriptor) descriptor1).isCache();
                boolean repo2IsCache = ((LocalRepoDescriptor) descriptor2).isCache();

                //Cache repositories should appear in a higher priority
                if (repo1IsCache && !repo2IsCache) {
                    return 1;
                } else if (!repo1IsCache && repo2IsCache) {
                    return -1;
                }
            }
            return descriptor1.getKey().compareTo(descriptor2.getKey());
        }
    }
}