/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.admin.configuration.repositories.replication;

import com.google.common.collect.Lists;
import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpStatus;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.replication.ReplicationAddon;
import org.artifactory.descriptor.replication.RemoteReplicationDescriptor;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.replication.remote.RemoteReplicationConfigModel;
import org.artifactory.ui.rest.service.admin.configuration.repositories.util.RemoteRepositoryTestUrl;
import org.artifactory.ui.rest.service.admin.configuration.repositories.util.builder.ReplicationConfigDescriptorBuilder;
import org.artifactory.ui.rest.service.admin.configuration.repositories.util.builder.RepoConfigDescriptorBuilder;
import org.artifactory.ui.rest.service.admin.configuration.repositories.util.validator.ReplicationConfigValidator;
import org.artifactory.ui.rest.service.admin.configuration.repositories.util.validator.RepoConfigValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.List;
import java.util.Optional;

import static org.apache.http.HttpStatus.SC_BAD_REQUEST;
import static org.apache.http.HttpStatus.SC_OK;
import static org.artifactory.ui.rest.service.admin.configuration.repositories.replication.TestLocalReplicationService.FAKE_CRON;

/**
 * @author Dan Feldman
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class TestRemoteReplicationService implements RestService<RemoteRepositoryConfigModel> {
    private static final Logger log = LoggerFactory.getLogger(TestRemoteReplicationService.class);

    @Autowired
    private AddonsManager addonsManager;

    @Autowired
    private ReplicationConfigDescriptorBuilder replicationDescriptorBuilder;

    @Autowired
    private RepoConfigValidator repoValidator;

    @Autowired
    private RepoConfigDescriptorBuilder repoDescriptorBuilder;

    @Autowired
    private ReplicationConfigValidator replicationValidator;

    @Autowired
    private RemoteRepositoryTestUrl remoteURlTester;

    @Override
    public void execute(ArtifactoryRestRequest<RemoteRepositoryConfigModel> request, RestResponse response) {
        RemoteRepositoryConfigModel model = request.getImodel();
        if (model == null) {
            response.error("No repository configuration given to test.").responseCode(SC_BAD_REQUEST);
            return;
        }
        HttpRepoDescriptor repo = null;
        try {
            setFakeCronExpForTestIfNeeded(model);
            repo = model.toDescriptor(repoValidator, repoDescriptorBuilder, null);
            RemoteReplicationDescriptor replication = model.getReplicationDescriptor(replicationValidator, replicationDescriptorBuilder);
            if (replication == null) {
                response.error("No Replication configuration was sent to test.").responseCode(HttpStatus.SC_NOT_FOUND);
                return;
            }
            remoteURlTester.testConnection(response, model);
            if (response.getResponseCode() != SC_OK) {
                return;
            }
            testRemoteReplication(replication, repo);
            response.info("Pull replication configuration tested successfully").responseCode(HttpStatus.SC_OK);
        } catch (Exception e) {
            log.debug("Error testing pull replication config: ", e);
            String error = "Error testing pull replication config: ";
            if (e instanceof UnknownHostException) {
                error += "\nUnknown host '";
                if (e.getMessage().equalsIgnoreCase("api") && repo != null) {
                    error += repo.getUrl();
                } else {
                    error += e.getMessage();
                }
            } else {
                error += e.getMessage();
            }
            response.error(error).responseCode(SC_BAD_REQUEST);
        }
    }

    private void testRemoteReplication(RemoteReplicationDescriptor replicationDescriptor, RemoteRepoDescriptor repo)
            throws IOException {
        if (StringUtils.isBlank(replicationDescriptor.getRepoKey())) {
            replicationDescriptor.setRepoKey(repo.getKey());
        }
        addonsManager.addonByType(ReplicationAddon.class).validateTargetIsDifferentInstance(replicationDescriptor, repo);
    }

    /**
     * Sets fake cron expressions in the replication descriptors just to allow for the test to run
     * if non were set by the ui
     */
    private void setFakeCronExpForTestIfNeeded(RemoteRepositoryConfigModel model) {
        List<RemoteReplicationConfigModel> replications =
                Optional.ofNullable(model.getReplications()).orElse(Lists.newArrayList());
        replications.stream()
                .filter(replication -> StringUtils.isBlank(replication.getCronExp()))
                .forEach(emptyReplication -> emptyReplication.setCronExp(FAKE_CRON));
    }
}
