/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.artifacts.search.packagesearch;

import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpStatus;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.metadata.service.MetadataEventService;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.security.auth.ActivePrincipalTokenStore;
import org.artifactory.ui.rest.model.artifacts.search.PackagesMetadataInfoModel;
import org.artifactory.ui.rest.model.artifacts.search.packagesearch.search.AqlUISearchModel;
import org.jfrog.metadata.client.MetadataClient;
import org.jfrog.metadata.client.model.MdsPackage;
import org.jfrog.metadata.client.model.MetadataAttribute;
import org.jfrog.metadata.client.model.SortTypeEnum;
import org.jfrog.metadata.client.rest.SearchResponse;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.Collections;
import java.util.List;
import java.util.Map;

/**
 * @author Elad Hirsch
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class PackagesMetadataSearchService implements RestService {

    private static final Logger log = LoggerFactory.getLogger(PackagesMetadataSearchService.class);

    private static final List<MetadataAttribute> PACKAGES_ATTRIBUTES = Lists
            .newArrayList(MetadataAttribute.LICENSES);

    @Autowired
    private MetadataEventService metadataEventService;

    @Autowired
    private ActivePrincipalTokenStore activePrincipalTokenStore;

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        try {
            MetadataClient metadataClient = metadataEventService.getMetadataClient();
            if (metadataClient != null) {
                SearchResponse<MdsPackage> searchResponse = metadataClient
                        .search()
                        .useActivePrincipalAuth(activePrincipalTokenStore.getOrLoadToken())
                        .packages(generateParams(request), PACKAGES_ATTRIBUTES,
                                SortTypeEnum.fromSql(request.getQueryParamByKey("order"))
                                , Collections.singletonList(request.getQueryParamByKey("sort_by"))
                                , StringUtils.isNotBlank(request.getQueryParamByKey("limit")) ?
                                        Integer.valueOf(request.getQueryParamByKey("limit")) : null);
                response.iModel(new PackagesMetadataInfoModel(searchResponse.getResults()));
            } else {
                log.warn("The Package-Native UI cannot be browsed while the Metadata Service is not available");
            }
            response.responseCode(HttpStatus.SC_OK);
        } catch (Exception e) {
            log.error("Packages metadata search error", e);
            response.responseCode(HttpStatus.SC_INTERNAL_SERVER_ERROR).error("Packages metadata search error");
        }
    }

    private Map<String, String> generateParams(ArtifactoryRestRequest request) {
        Map<String, String> params = Maps.newHashMap();
        String repoType = RepoType.Docker.getType().equals(request.getPathParamByKey("type")) ? "docker" : "npm";
        params.put("package_type", repoType);


        List<AqlUISearchModel> searches = request.getModels();
        if (searches != null) {
            searches.forEach(attribute -> {
                String id = attribute.getId();
                String value = attribute.getValues().get(0);

                if (id.equals("pkg") || id.equals("npmName")) {
                    params.put("name", value);
                }

                if (id.equals("npmVersion") || id.equals("version")) {
                    params.put("versions.name", value);
                }

                if (id.equals("npmKeywords")) {
                    params.put("tags.value", value);
                }

                if (id.equals("npmScope")) {
                    params.put("qualifiers.name", "npm.scope");
                    params.put("qualifiers.value", value);
                }

                if (id.equals("npmChecksum")) {
                    if (value.length() == 32) {
                        params.put("versions.files.md5", value);
                    } else if (value.length() == 40) {
                        params.put("versions.files.sha1", value);
                    } else {
                        params.put("versions.files.sha256", value);
                    }
                }
            });
        }

        return params;
    }


}
