/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.rest.services.permissions;

import com.google.common.base.Joiner;
import com.google.common.collect.Lists;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.CoreAddons;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.rest.constant.RestConstants;
import org.artifactory.api.security.AclService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.build.InternalBuildService;
import org.artifactory.factory.InfoFactoryHolder;
import org.artifactory.model.xstream.security.BuildPermissionTargetImpl;
import org.artifactory.model.xstream.security.MutableRepoAclImpl;
import org.artifactory.model.xstream.security.RepoPermissionTargetImpl;
import org.artifactory.rest.common.security.SecurityModelPopulator;
import org.artifactory.rest.exception.BadRequestException;
import org.artifactory.rest.exception.ConflictException;
import org.artifactory.rest.exception.ForbiddenWebAppException;
import org.artifactory.rest.exception.NotFoundException;
import org.artifactory.security.*;
import org.artifactory.security.permissions.*;
import org.artifactory.storage.StorageException;
import org.artifactory.util.HttpUtils;
import org.jfrog.common.ClockUtils;
import org.jfrog.common.StreamSupportUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.userdetails.UsernameNotFoundException;
import org.springframework.stereotype.Component;

import javax.annotation.Nullable;
import javax.servlet.http.HttpServletRequest;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.apache.commons.collections.CollectionUtils.isNotEmpty;
import static org.apache.commons.lang.StringUtils.isNotBlank;
import static org.artifactory.api.rest.constant.SecurityRestConstants.PATH_ROOT_V2;
import static org.artifactory.api.rest.constant.SecurityRestConstants.PERMISSIONS_ROOT;
import static org.artifactory.rest.services.permissions.RestSecurityHelperCommon.*;
import static org.artifactory.security.ArtifactoryResourceType.BUILD;
import static org.artifactory.security.ArtifactoryResourceType.REPO;
import static org.artifactory.security.PermissionTargetNaming.NAMING_DISPLAY;
import static org.artifactory.security.PermissionTargetNaming.NAMING_UI;

/**
 * Common methods for CRUD actions on permission targets via REST api and ui.
 *
 * @author Yuval Reches
 * @author Dan Feldman
 */
@Component
public class RestSecurityRequestHandlerV2Impl implements InternalRestSecurityRequestHandlerV2 {
    private static final Logger log = LoggerFactory.getLogger(RestSecurityRequestHandlerV2Impl.class);

    public static final String MISSING_PARTS_ERR_MSG = "Permission target request missing repositories and builds. Must have at least one.";
    public static final String ALREADY_EXISTS_ERR_MSG = "Can't create permission target '%s' for type %s. It already exists.";

    private static final String USERS = "users";
    private static final String GROUPS = "groups";
    private AclService aclService;
    private UserGroupService userGroupService;
    private CentralConfigService configService;
    private AuthorizationService authorizationService;
    private InternalBuildService buildService;
    private AddonsManager addonsManager;

    @Autowired
    public RestSecurityRequestHandlerV2Impl(AclService aclService,
            UserGroupService userGroupService, CentralConfigService configService,
            AuthorizationService authorizationService, InternalBuildService buildService,
            AddonsManager addonsManager) {
        this.aclService = aclService;
        this.userGroupService = userGroupService;
        this.configService = configService;
        this.authorizationService = authorizationService;
        this.buildService = buildService;
        this.addonsManager = addonsManager;
    }

    /*
     * Public methods
     */

    @Override
    public boolean isPermissionTargetExists(String entityKey) throws ForbiddenWebAppException, NotFoundException {
        Pair<RepoAcl, BuildAcl> permissionExist = getExistingPermissionTarget(aclService, entityKey);
        RepoAcl repoAcl = permissionExist.getLeft();
        BuildAcl buildAcl = permissionExist.getRight();

        // Validate user can get the permission
        boolean repoExistsAndDenied = isAclExistsAndDenied(repoAcl);
        boolean buildExistsAndDenied = isAclExistsAndDenied(buildAcl);
        validateCanManagePermission(entityKey, repoExistsAndDenied, buildExistsAndDenied, repoAcl == null, buildAcl == null);

        // Removing the acls from response in case no manage permission
        repoAcl = repoExistsAndDenied ? null : repoAcl;
        buildAcl = buildExistsAndDenied ? null : buildAcl;

        return repoAcl != null || buildAcl != null;
    }

    @Override
    public List<PermissionTargetModel> getAllPermissionTargets(PermissionTargetNaming outputMode) {
        Map<String, RepoAcl> keyToRepoAcl = aclService.getAllRepoAcls(ArtifactoryPermission.MANAGE).stream()
                .collect(Collectors.toMap(repoAcl -> repoAcl.getPermissionTarget().getName(), repoAcl -> repoAcl));

        Map<String, BuildAcl> keyToBuildAcl = aclService.getAllBuildAcls(ArtifactoryPermission.MANAGE).stream()
                .collect(Collectors.toMap(buildAcl -> buildAcl.getPermissionTarget().getName(), buildAcl -> buildAcl));

        //Aggregate same-name repo and build acls into one model, since we got only user-manageable permissions no need to validate
        //Always use UI naming convention since this method's only user is the UI.
        List<PermissionTargetModel> list = collectRepoAndBuildPermissions(keyToRepoAcl, keyToBuildAcl, outputMode);
        // Adding all the build acls that did not mach any repo acl name
        keyToBuildAcl.keySet().stream()
                .map(aclKey -> toPermissionTargetModel(null, keyToBuildAcl.get(aclKey), outputMode))
                .forEach(list::add);
        return list;
    }

    private List<PermissionTargetModel> collectRepoAndBuildPermissions(Map<String, RepoAcl> keyToRepoAcl,
            Map<String, BuildAcl> keyToBuildAcl, PermissionTargetNaming outputMode) {
        return keyToRepoAcl.keySet().stream()
                .map(aclKey -> toPermissionTargetModel(keyToRepoAcl.get(aclKey), keyToBuildAcl.get(aclKey), outputMode))
                .peek(permission -> keyToBuildAcl.remove(permission.getName()))
                .collect(Collectors.toList());
    }

    @Override
    public Set<SecurityEntityListItem> getAllPermissionTargets(HttpServletRequest request) {
        return getAllPermissionTargets(NAMING_DISPLAY).stream()
                .map(permission -> getPermissionListItem(request, permission))
                .collect(Collectors.toSet());
    }

    private SecurityEntityListItem getPermissionListItem(HttpServletRequest request, PermissionTargetModel permission) {
        String name = permission.getName();
        return new SecurityEntityListItem(name, generatePermissionUri(request, name));
    }

    private static String generatePermissionUri(HttpServletRequest request, String permissionName) {
        String servletContextUrl = HttpUtils.getServletContextUrl(request);
        return Joiner.on("/").join(servletContextUrl, RestConstants.PATH_API, PATH_ROOT_V2, PERMISSIONS_ROOT,
                HttpUtils.encodeQuery(permissionName));
    }

    @Override
    public PermissionTargetModel getPermissionTarget(String entityKey, PermissionTargetNaming outputMode) throws ForbiddenWebAppException, NotFoundException  {
        RepoAcl repoAcl = aclService.getRepoAcl(entityKey);
        BuildAcl buildAcl = aclService.getBuildAcl(entityKey);
        if (repoAcl == null && buildAcl == null) {
            return null;
        }

        // Validate user can get the permission
        boolean repoExistsAndDenied = isAclExistsAndDenied(repoAcl);
        boolean buildExistsAndDenied = isAclExistsAndDenied(buildAcl);
        validateCanManagePermission(entityKey, repoExistsAndDenied, buildExistsAndDenied,repoAcl == null, buildAcl == null);

        // Removing the acls from response in case no manage permission
        log.debug("Checking for existence of acl entities and removing them from response if needed");
        repoAcl = repoExistsAndDenied ? null : repoAcl;
        buildAcl = buildExistsAndDenied ? null : buildAcl;

        return toPermissionTargetModel(repoAcl, buildAcl, outputMode);
    }

    private PermissionTargetModel toPermissionTargetModel(RepoAcl repoAcl, BuildAcl buildAcl, PermissionTargetNaming outputMode ) {
        if (repoAcl != null) {
            MutableRepoAclImpl toModifyAcl = new MutableRepoAclImpl(repoAcl);
            repoAcl = aclService.convertNewAclCachedRepoKeysToRemote(toModifyAcl);
        }
        return SecurityModelPopulator.getPermissionTargetModelV2(repoAcl, buildAcl, outputMode);
    }

    @Override
    public void createPermissionTarget(String entityKey, PermissionTargetModel permissionTargetModel, PermissionTargetNaming inputMode) {
        if (!authorizationService.isAdmin()) {
            throw new ForbiddenWebAppException("Only and admin user can create permission targets.");
        }
        validateIncomingPermissionTarget(entityKey, permissionTargetModel);

        // In case permission already exists we throw error
        validatePermissionsExistenceForCreate(entityKey, permissionTargetModel);
        // Validate the user input is valid
        validateRepoAndBuildSections(permissionTargetModel, inputMode, null);

        Acl repoAcl = createRepoAclModel(entityKey, permissionTargetModel, null, inputMode);
        Acl buildAcl = createBuildAclModel(entityKey, permissionTargetModel, null, inputMode);
        // Create the ACLs in access
        getTxMe().createAcl(repoAcl, buildAcl);
    }

    @Override
    public void updatePermissionTarget(String entityKey, PermissionTargetModel permissionTargetModel, PermissionTargetNaming inputMode) {
        validateIncomingPermissionTarget(entityKey, permissionTargetModel);
        Pair<RepoAcl, BuildAcl> existingPermission = getExistingPermissionTarget(aclService, entityKey);
        RepoAcl existingRepoAcl = existingPermission.getLeft();
        BuildAcl existingBuildAcl = existingPermission.getRight();

        //What permission targets came in with the model?
        boolean hasRepoPermissionInRequest = hasRepoPermission(permissionTargetModel);
        boolean hasBuildPermissionInRequest = hasBuildPermission(permissionTargetModel);

        validateUpdateActionPermissions(entityKey, existingRepoAcl, existingBuildAcl, hasRepoPermissionInRequest, hasBuildPermissionInRequest);

        //Validate incoming request for discrepancies in model
        validateRepoAndBuildSections(permissionTargetModel, inputMode, existingPermission);

        Acl repoAcl = createRepoAclModel(entityKey, permissionTargetModel, existingRepoAcl, inputMode);
        Acl buildAcl = createBuildAclModel(entityKey, permissionTargetModel, existingBuildAcl, inputMode);

        boolean repoAclExists = existingRepoAcl != null;
        boolean buildAclExists = existingBuildAcl != null;

        List<Acl> toAdd = getAclAccordingToExistence(repoAcl, buildAcl, !repoAclExists, !buildAclExists);
        List<Acl> toUpdate = getAclAccordingToExistence(repoAcl, buildAcl, repoAclExists, buildAclExists);
        List<Acl> toDelete = getAclsToDeleteIfNeeded(entityKey, existingRepoAcl, existingBuildAcl, repoAclExists,
                buildAclExists, hasRepoPermissionInRequest, hasBuildPermissionInRequest);

        getTxMe().updateAcl(toUpdate, toDelete, toAdd);
    }

    /**
     * A set of validations on what can the user actually change based on what came in with the response.
     * A second set of validations on the content itself is performed later inside {@link #createRepoAclModel}
     * and {@link #createBuildAclModel}
     */
    private void validateUpdateActionPermissions(String entityKey, RepoAcl existingRepoAcl, BuildAcl existingBuildAcl,
            boolean hasRepoPermissionInRequest, boolean hasBuildPermissionInRequest) {
        //What can this user change?
        boolean repoExistsAndDenied = isAclExistsAndDenied(existingRepoAcl);
        boolean buildExistsAndDenied = isAclExistsAndDenied(existingBuildAcl);

        //Validate can manage on any of them
        //validateCanManagePermission(entityKey, repoExistsAndDenied, buildExistsAndDenied, repoAclExists, buildAclExists);

        //If any model that the user can't manage came in with the request, fail it.
        if (repoExistsAndDenied && hasRepoPermissionInRequest) {
            throwForbiddenPermissionErrorMessage(entityKey, getDeniedByType(true));
        } else if (buildExistsAndDenied && hasBuildPermissionInRequest) {
            throwForbiddenPermissionErrorMessage(entityKey, getDeniedByType(false));
        }
        // At this point user can either manage both and both exists, or the one they can't manage isn't present anyway
    }

    //On the update list we add to the list if the acl already exists, on the create list we add only if it does not
    private ArrayList<Acl> getAclAccordingToExistence(Acl repoAcl, Acl buildAcl, boolean shouldAddRepoToList, boolean shouldAddBuildToList) {
        ArrayList<Acl> aclList = Lists.newArrayList();
        if (shouldAddRepoToList && repoAcl != null) {
            aclList.add(repoAcl);
        }
        if (shouldAddBuildToList && buildAcl != null) {
            aclList.add(buildAcl);
        }
        return aclList;
    }

    @Override
    public void deletePermissionTarget(String entityKey) {
        Pair<RepoAcl, BuildAcl> existingPermission = getExistingPermissionTarget(aclService, entityKey);
        RepoAcl repoAcl = existingPermission.getLeft();
        BuildAcl buildAcl = existingPermission.getRight();
        boolean repoAclMissing = repoAcl == null;
        boolean buildAclMissing = buildAcl == null;
        if (repoAclMissing && buildAclMissing) {
            throw new NotFoundException("No such permission target '" + entityKey + "'");
        }
        //nulls filtered by underlying call
        getTxMe().deleteAcl(repoAcl, buildAcl);
    }

    @Override
    public List<String> getBuildsPerPatterns(List<String> includePatterns, List<String> excludePatterns) {
        return getAllBuildsInPermission(includePatterns, excludePatterns);
    }

    @Override
    public List<SecurityEntityPermissionTargetModel> getUserPermissionSecurityEntity(String userName) {
        validateUser(userName);
        return SecurityModelPopulator.getSecurityEntityPermissionTargetModelForRest(getUserRepoPermission(userName, REPO),
                getUserRepoPermission(userName, BUILD));
    }

    @Override
    public List<SecurityEntityPermissionTargetModel> getGroupPermissionSecurityEntity(String groupName) {
        GroupInfo groupInfo = userGroupService.findGroup(groupName);
        if (groupInfo == null) {
            throw new NotFoundException();
        }
        return SecurityModelPopulator.getSecurityEntityPermissionTargetModelForRest(getGroupRepoPermission(groupName, REPO),
                getGroupRepoPermission(groupName, BUILD));
    }

    private List<String> getAllBuildsInPermission(List<String> includePatterns, List<String> excludePatterns) {
        return buildService.getBuildNamesInternally().stream()
                .filter(build -> userGroupService.isBuildInPermissions(includePatterns, excludePatterns, build))
                .collect(Collectors.toList());
    }

    /*
     * Model
     */

    /**
     * In case Repositories section in user request is empty --> return null
     */
    @Nullable
    private Acl createRepoAclModel(String entityKey, PermissionTargetModel permissionTargetModel, @Nullable RepoAcl previousAcl, PermissionTargetNaming inputMode) {
        if (!hasRepoPermission(permissionTargetModel)) {
            return null;
        }
        validateRepoAclModifications(previousAcl, permissionTargetModel.getRepo().getRepositories());
        String currentUsername = authorizationService.currentUsername();
        // Permission Target
        RepoPermissionTargetImpl repoPermissionTarget = new RepoPermissionTargetImpl(entityKey);
        RepoPermissionTargetModel repositories = permissionTargetModel.getRepo();
        repoPermissionTarget.setRepoKeys(repositories.getRepositories());
        repoPermissionTarget.setIncludes(repositories.getIncludePatterns());
        repoPermissionTarget.setExcludes(repositories.getExcludePatterns());
        // Get Aces and perform user request validation for actions
        Set<AceInfo> aceInfos = getAcesForAclAndValidateActions(inputMode, repositories.getActions());
        if (previousAcl != null) {
            return InfoFactoryHolder.get()
                    .createRepoAcl(repoPermissionTarget, aceInfos, currentUsername,
                            ClockUtils.epochMillis(), previousAcl.getAccessIdentifier());
        }
        return InfoFactoryHolder.get()
                .createRepoAcl(repoPermissionTarget, aceInfos, currentUsername,
                        ClockUtils.epochMillis());
    }

    /**
     * If user is not admin they can only touch the patterns and principals
     * @param previousAcl  Previous value of the acl being modified, if exists.
     * @param newRepoKeys  New permission target's repos about to be sent to the update action.
     */
    private void validateRepoAclModifications(@Nullable RepoAcl previousAcl, List<String> newRepoKeys) {
        //Manage permission already verified at this point
        if (previousAcl != null && !authorizationService.isAdmin()) {
            diffListAndError(aclService.convertCachedRepoKeysToRemote(newRepoKeys),
                    aclService.convertCachedRepoKeysToRemote(previousAcl.getPermissionTarget().getRepoKeys()), "repositories");
        }
    }

    /**
     * Try to get Ace set for Acl and validate that the user request is with valid permissions.
     * @return a set of converted Aces into internal naming
     * @throws BadRequestException in case the user request contains invalid permissions
     */
    private Set<AceInfo> getAcesForAclAndValidateActions(PermissionTargetNaming inputMode, PrincipalConfiguration principals) {
        try {
            return getAcesForAcl(principals, inputMode);
        } catch (IllegalArgumentException iae) {
            throw new BadRequestException(iae.getMessage());
        }
    }

    /**
     * In case Builds section in user request is empty --> return null
     */
    @Nullable
    private Acl createBuildAclModel(String entityKey, PermissionTargetModel permissionTargetModel, @Nullable BuildAcl previousAcl, PermissionTargetNaming inputMode) {
        if (!hasBuildPermission(permissionTargetModel)) {
            return null;
        }
        validateBuildAclModifications(previousAcl, permissionTargetModel.getBuild());
        String currentUsername = authorizationService.currentUsername();
        // Permission Target
        BuildPermissionTargetImpl buildPermissionTarget = new BuildPermissionTargetImpl(entityKey);
        RepoPermissionTargetModel builds = permissionTargetModel.getBuild();
        Set<AceInfo> aceInfos = setPermissionTargetFieldsAndGetAces(inputMode, buildPermissionTarget, builds);
        if (previousAcl != null) {
            return InfoFactoryHolder.get()
                    .createBuildAcl(buildPermissionTarget, aceInfos, currentUsername,
                            ClockUtils.epochMillis(), previousAcl.getAccessIdentifier());
        }
        return InfoFactoryHolder.get()
                .createBuildAcl(buildPermissionTarget, aceInfos, currentUsername,
                        ClockUtils.epochMillis());
    }

    /**
     * If the user is not admin they can only touch principals
     * @param previousAcl     Previous value of the acl being modified, if exists.
     * @param newPermission   New incoming permission target to validate
     */
    private void validateBuildAclModifications(@Nullable BuildAcl previousAcl, RepoPermissionTargetModel newPermission) {
        if (previousAcl != null && !authorizationService.isAdmin()) {
            BuildPermissionTarget previousPermission = previousAcl.getPermissionTarget();
            diffListAndError(newPermission.getIncludePatterns(), previousPermission.getIncludes(), "include patterns");
            diffListAndError(newPermission.getExcludePatterns(), previousPermission.getExcludes(), "exclude patterns");
        }
    }

    private void diffListAndError(List<String> newElementsList, List<String> oldElementsList, String element) {
        //Create a copy so we don't mess up the model, its still being used.
        List<String> newElements = new ArrayList<>(newElementsList);
        newElements.removeAll(oldElementsList);
        if(isNotEmpty(newElements)) {
            throw new ForbiddenWebAppException("Manager user cannot modify " + element + " in a permission target");
        }
    }

    private Set<AceInfo> setPermissionTargetFieldsAndGetAces(PermissionTargetNaming inputMode,
            BuildPermissionTargetImpl buildPermissionTarget, RepoPermissionTargetModel builds) {
        buildPermissionTarget.setRepoKeys(builds.getRepositories());
        buildPermissionTarget.setIncludes(builds.getIncludePatterns());
        buildPermissionTarget.setExcludes(builds.getExcludePatterns());
        // Get Aces and perform user request validation for actions
        return getAcesForAclAndValidateActions(inputMode, builds.getActions());
    }

    /*
     * Actions
     */

    /**
     * @throws BadRequestException on error creating an acl.
     */
    public void createAcl(Acl... acls) {
        try {
            Stream.of(acls)
                    .filter(Objects::nonNull)
                    .forEach(acl -> aclService.createAcl(acl));
        } catch (StorageException e) {
            log.warn("Failed to create permission: {}", e.getCause().getMessage());
            log.debug("Failed to create permission: {}", e);
            throw new BadRequestException("Failed to create permission: " + e.getCause().getMessage());
        }
    }

    /**
     * @param toUpdate - acls to update
     * @param toDelete - to maintain atomicity this is also the correct place to delete any acls if a request contained
     *                   null value for a pre-existing acl while it contained another one to update.
     * @param toAdd    - acls to add
     * @throws BadRequestException on error creating an acl.
     */
    public void updateAcl(List<Acl> toUpdate, List<Acl> toDelete, List<Acl> toAdd) throws BadRequestException {
        try {
            toDelete.stream()
                    .filter(Objects::nonNull)
                    .forEach(aclService::deleteAcl);
            toUpdate.stream()
                    .filter(Objects::nonNull)
                    .forEach(aclService::updateAcl);
            toAdd.stream()
                    .filter(Objects::nonNull)
                    .forEach(aclService::createAcl);
        } catch (StorageException e) {
            log.warn("Failed to update permission: {}", e.getCause().getMessage());
            log.debug("Failed to update permission: {}", e);
            throw new BadRequestException("Failed to update permission: " + e.getCause().getMessage());
        }
    }

    /**
     * @throws BadRequestException on error creating an acl.
     */
    public void deleteAcl(Acl... acls) throws BadRequestException {
        try {
            Stream.of(acls)
                    .filter(Objects::nonNull)
                    .forEach(acl -> aclService.deleteAcl(acl));
        } catch (StorageException e) {
            log.warn("Failed to delete permission: {}", e.getCause().getMessage());
            log.debug("Failed to delete permission: {}", e);
            throw new BadRequestException("Failed to delete permission: " + e.getCause().getMessage());
        }
    }

    /**
     * Request is treated as delete of old ACLs in case the section is missing in the model and the ACL exists,
     * but treat as actual delete only if the user has manage on it
     *
     * @return the acls to delete if required, they will be passed to the {@link #updateAcl} method to maintain atomicity
     */
    private List<Acl> getAclsToDeleteIfNeeded(String entityKey, RepoAcl existingRepoAcl, BuildAcl existingBuildAcl,
            boolean repoAclExists, boolean buildAclExists, boolean hasRepoPermissionInRequest, boolean hasBuildPermissionInRequest) {
        List<Acl> toDelete = Lists.newArrayList();

        if (repoAclExists && !hasRepoPermissionInRequest && aclService.canManage(existingRepoAcl.getPermissionTarget())) {
            log.debug("Will delete permission {} of type repo as its missing in user request but exists in DB", entityKey);
            toDelete.add(existingRepoAcl);
        }
        if (buildAclExists && !hasBuildPermissionInRequest && aclService.canManage(existingBuildAcl.getPermissionTarget())) {
            log.debug("Will delete permission {} of type build as its missing in user request but exists in DB", entityKey);
            toDelete.add(existingBuildAcl);
        }
        return toDelete;
    }

    private InternalRestSecurityRequestHandlerV2 getTxMe() {
        return ContextHelper.get().beanForType(InternalRestSecurityRequestHandlerV2.class);
    }

    /*
     * Assertions and validations
     */

    /**
     *
     * @throws ConflictException    on entity name mismatch
     * @throws BadRequestException  on no entities given
     */
    private void validateIncomingPermissionTarget(String entityKey, PermissionTargetModel permissionTargetModel) {
        String name = permissionTargetModel.getName();
        if (isNotBlank(name) && !entityKey.equals(name)) {
            throw new ConflictException(CONFLICT_ERR_MSG);
        }
        if (!hasRepoPermission(permissionTargetModel) && !hasBuildPermission(permissionTargetModel)) {
            throw new BadRequestException(MISSING_PARTS_ERR_MSG);
        }
    }

    private void validateRepoAndBuildSections(PermissionTargetModel permissionTargetModel, PermissionTargetNaming inputMode,
                                              Pair<RepoAcl, BuildAcl> existingPermission) {
        validateRepositoriesSection(permissionTargetModel, existingPermission);
        validateBuildsSection(permissionTargetModel, inputMode, existingPermission);
    }

    /**
     * Validate that we don't already have the ACLs with same entity key
     *
     * @throws BadRequestException in case of assertion error
     */
    private void validatePermissionsExistenceForCreate(String entityKey, PermissionTargetModel permissionTargetModel) {
        Pair<RepoAcl, BuildAcl> permissionExist = getExistingPermissionTarget(aclService, entityKey);
        boolean repoAclExists = permissionExist.getLeft() != null;
        boolean buildAclExists = permissionExist.getRight() != null;
        if (repoAclExists && hasRepoPermission(permissionTargetModel)) {
            generatePermissionExistsError(entityKey, ArtifactoryResourceType.REPO.getName());
        }
        if (buildAclExists && hasBuildPermission(permissionTargetModel)) {
            generatePermissionExistsError(entityKey, ArtifactoryResourceType.BUILD.getName());
        }
    }

    /**
     * Requires at least one build name OR build regex. Can't have both.
     * Asserting all build names exist, principals exist.
     *
     * @throws BadRequestException in case of assertion errors
     */
    private void validateBuildsSection(PermissionTargetModel permissionTargetModel,
            PermissionTargetNaming inputMode, Pair<RepoAcl, BuildAcl> existingPermission) throws BadRequestException {
        RepoPermissionTargetModel builds = permissionTargetModel.getBuild();
        if (builds == null) {
            log.debug("No build permissions content in request for Acl {} of type {}", permissionTargetModel.getName(), BUILD.getName());
            return;
        }
        // Asserting common part between repo and build
        validateRepoPermissionModel(builds, existingPermission);
        // Repo key validation
        assertBuildModelRequiredEntities(permissionTargetModel, inputMode);
    }

    /**
     * Requires repositories list and valid principals
     *
     * @throws BadRequestException in case of assertion error
     */
    private void validateRepositoriesSection(PermissionTargetModel permissionTargetModel, Pair<RepoAcl, BuildAcl> existingPermission)
            throws BadRequestException {
        RepoPermissionTargetModel repositories = permissionTargetModel.getRepo();
        if (repositories == null) {
            log.debug("No repo permissions content in request for Acl {} of type {}", permissionTargetModel.getName(), REPO.getName());
            return;
        }
        validateRepoPermissionModel(repositories, existingPermission);
        if (repositories.getRepositories().contains(buildService.getBuildInfoRepoKey())) {
            throw new BadRequestException("Setting repository permissions on Build Info Repository is not allowed.");
        }
    }

    private void validateRepoPermissionModel(RepoPermissionTargetModel repositories, Pair<RepoAcl, BuildAcl> existingPermission) {
        if (repositories.getRepositories() == null) {
            throw new BadRequestException("Permission target request missing repositories.");
        }
        // RepoKeys validation
        String missingRepo = getFirstNonExistingRepoFromList(repositories.getRepositories(), configService, aclService);
        if (isNotBlank(missingRepo)) {
            throw new BadRequestException("Permission target contains a reference to a non-existing repository '" + missingRepo + "'.");
        }
        // Principals validation
        checkForNonExistingPrinciples(userGroupService, repositories.getActions(), existingPermission);
    }

    /**
     * Requires that {@link InternalBuildService#getBuildInfoRepoKey()} is the only repo key in the repositories list.
     * In case request came from UI we add the repo ourselves. Otherwise throw exception.
     *
     * @throws BadRequestException on assertion error.
     */
    private void assertBuildModelRequiredEntities(PermissionTargetModel permissionTargetModel,
            PermissionTargetNaming inputMode) throws BadRequestException {
        RepoPermissionTargetModel builds = permissionTargetModel.getBuild();
        List<String> buildInfoRepoList = Collections.singletonList(buildService.getBuildInfoRepoKey());
        if (!buildInfoRepoList.equals(builds.getRepositories())) {
            if (NAMING_UI.equals(inputMode)) {
                builds.setRepositories(buildInfoRepoList);
            } else {
                throw new BadRequestException(
                        "Build permission target repositories list must contain " + buildService.getBuildInfoRepoKey() +
                                " as the only repository key");
            }
        }
    }

    /**
     * Throws when user doesn't have manage on both permission targets, for actions that require manage on at least one
     * of them.
     *
     * @throws ForbiddenWebAppException if permission was denied for {@param entityKey}
     * @throws NotFoundException  if permission was denied for {@param entityKey} and Artifactory is set to hide existence
     */
    private void validateCanManagePermission(String entityKey, boolean repoExistsAndDenied, boolean buildExistsAndDenied, boolean repoMissing, boolean buildMissing) {
        if (aclsForbiddenForUser(repoExistsAndDenied, buildExistsAndDenied, repoMissing, buildMissing)) {
            throwForbiddenPermissionErrorMessage(entityKey, getDeniedPermissionType(repoExistsAndDenied, buildExistsAndDenied));
        }
    }

    private boolean aclsForbiddenForUser(boolean repoExistsAndDenied, boolean buildExistsAndDenied, boolean repoMissing, boolean buildMissing) {
        return (repoExistsAndDenied && buildExistsAndDenied) ||
                (repoExistsAndDenied && buildMissing) ||
                (buildExistsAndDenied && repoMissing);
    }

    private boolean isAclExistsAndDenied(Acl<? extends PermissionTarget> acl) {
        return acl != null && !aclService.canManage(acl.getPermissionTarget());
    }

    private boolean hasRepoPermission(PermissionTargetModel model) {
        return model != null && model.getRepo() != null;
    }

    private boolean hasBuildPermission(PermissionTargetModel model) {
        return model != null && model.getBuild() != null;
    }

    /*
     * Errors
     */

    private void generatePermissionExistsError(String entityKey, String type) {
        throw new ConflictException(String.format(ALREADY_EXISTS_ERR_MSG, entityKey, type));
    }

    /**
     * @throws ForbiddenWebAppException if permission was denied for {@param entityKey}
     * @throws NotFoundException if permission was denied for {@param entityKey} and Artifactory is set to hide existence
     * of unauthorized resources.
     */
    private void throwForbiddenPermissionErrorMessage(String entityKey, String deniedType) {
        if (configService.getDescriptor().getSecurity().isHideUnauthorizedResources()) {
            log.debug("User tried to access permission {} but it is forbidden and " +
                    "isHideUnauthorizedResources is true. Returning 404 instead of 403.", entityKey, deniedType);
            throw new NotFoundException("");
        }
        throw new ForbiddenWebAppException("Accessing permission target '" + entityKey + "' " + deniedType + "is forbidden.");
    }

    private String getDeniedPermissionType(boolean repoExistsAndDenied, boolean buildExistsAndDenied) {
        return repoExistsAndDenied && buildExistsAndDenied ? StringUtils.EMPTY : getDeniedByType(repoExistsAndDenied);
    }

    private String getDeniedByType(boolean repoExistsAndDenied) {
        return repoExistsAndDenied ?
                    "of type " + ArtifactoryResourceType.REPO.getName() + " " :
                    "of type " + ArtifactoryResourceType.BUILD.getName() + " ";
    }

    private Map<String, SecurityEntityRepoPermissionTargetModel> getGroupRepoPermission(String groupName,
            ArtifactoryResourceType type) {
        List<String> groups = Collections.singletonList(groupName);
        return StreamSupportUtils.multimapEntriesStream(aclService.getGroupsPermissions(groups, type))
                .collect(Collectors.toMap(entry -> entry.getKey().getName(),
                        entry -> SecurityModelPopulator.getSecurityEntityRepoPermissionTargetModelV2(entry.getValue(),
                                (RepoPermissionTarget)  entry.getKey())));

    }

    private Map<String, SecurityEntityRepoPermissionTargetModel> getUserRepoPermission(String userName,
            ArtifactoryResourceType type) {
        return StreamSupportUtils.mapEntriesStream(aclService.getUserPermissionAndItsGroups(userName, type).asMap())
                .collect(Collectors.toMap(entry -> entry.getKey().getName(),
                        entry -> SecurityModelPopulator.getSecurityEntityRepoPermissionTargetModelV2(entry.getValue(),
                                (RepoPermissionTarget)  entry.getKey())));
    }

    private void validateUser(String userName) {
        try {
            UserInfo user = userGroupService.findUser(userName);
            if (addonsManager.addonByType(CoreAddons.class).isAolAdmin(user)) {
                throw new NotFoundException();
            }
        }
        catch (UsernameNotFoundException e) {
            throw new NotFoundException();
        }
    }

}
